import logging
import os
import tempfile
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

import numpy as np
import onnxruntime as ort

from highlighter.core.cuda import resolve_device_id
from highlighter.io.url import is_url_scheme
from highlighter.predictors.ort.session import get_ort_options, get_session
from highlighter.predictors.ort.utils import download_model
from highlighter.predictors.ort.utils import is_cuda_available as _is_cuda_available

HL_GPU_DEVICE_ID = resolve_device_id()
HL_MODEL_HEAD_INDEX = os.environ.get("HL_MODEL_HEAD_INDEX", 0)
LOGGER = logging.getLogger("OnnxPredictor")
ort.set_default_logger_severity(os.environ.get("ONNX_RUNTIME_LOGLEVEL", 3))


def is_cuda_available():
    return _is_cuda_available()


class OnnxPredictor:
    """Loads an onnx model and does prediction

    Set log level:
       set the ONNX_RUNTIME_LOGLEVEL environment variable

    Set gpu id
       set the HL_GPU_DEVICE_ID environment variable

    Select a different head when reading TrainingRunInterface.model_outputs
       set the HL_MODEL_HEAD_INDEX environment variable
    """

    def __init__(
        self,
        onnx_file: Union[str, Path],
        device_id: Optional[int] = None,
        ort_custom_op_path: Optional[str] = None,
        artefact_cache_dir: Optional[Path] = None,
        onnx_file_download_timeout: int = 60,
    ):
        self._run_options = get_ort_options()
        if device_id is None:
            device_id = resolve_device_id()

        providers = ["CPUExecutionProvider"]
        provider_options = [{}]
        self.is_cuda_available = is_cuda_available()
        if self.is_cuda_available:
            providers.insert(0, "CUDAExecutionProvider")
            provider_options.insert(0, {"device_id": device_id})
        LOGGER.debug(f"is_cuda_available={self.is_cuda_available}")
        LOGGER.debug(f"provider_options={provider_options}")

        if is_url_scheme(str(onnx_file), ["http", "https"]):
            artefact_cache_dir = (
                Path(tempfile.mkdtemp()) if artefact_cache_dir is None else artefact_cache_dir
            )
            parsed_url = urlparse(str(onnx_file))
            onnx_path = artefact_cache_dir / parsed_url.path.strip("/")
            onnx_path.parent.mkdir(exist_ok=True, parents=True)
            download_model(
                parsed_url.geturl(),
                onnx_path,
                check_cached=True,
                timeout=onnx_file_download_timeout,
            )
        elif Path(onnx_file).exists():
            onnx_path = onnx_file
        else:
            raise ValueError(f"Unable to find onnx file at {onnx_file}")

        self.sess = get_session(
            str(onnx_path),
            providers=providers,
            provider_options=provider_options,
            custom_ops_path=ort_custom_op_path,
        )

        self.io_binding = self.sess.io_binding()

        _inputs = self.sess.get_inputs()
        if len(_inputs) > 1:
            raise ValueError(f"Expected only 1 input name for onnx-runtime session, got: {_inputs}")
        self.input_name = _inputs[0].name
        input_shape = _inputs[0].shape

        # 3D input probably indicates an ONNX model that expects a single image
        if len(input_shape) == 3:
            raise ValueError(
                f"Onnx models must have batch dimension. This Onnx model has an input of shape {input_shape}"
            )
        self.batch_size = input_shape[0]
        if isinstance(self.batch_size, int) and self.batch_size != 1:
            raise ValueError(
                f"Batch size must be the integer 1 or a symbolic value to indicate dynamic match size, got: {type(self.batch_size).__name__}: {self.batch_size}"
            )

        self.output_names = [_.name for _ in self.sess.get_outputs()]
        LOGGER.debug(f"onnxruntime session instantiated, input_shape {input_shape}")

    def predict(self, input_batch):
        if self.batch_size == 1:
            outputs = []
            for single_input in input_batch:
                single_input = np.expand_dims(single_input, axis=0)
                outputs.append(self.predict_batch(single_input)[0])

        elif not isinstance(self.batch_size, int):  # Symbolic value indicates dynamic batch size
            outputs = self.predict_batch(input_batch)
        else:
            raise ValueError(f"Unhandled batch size: {self.batch_size}")
        return outputs

    def predict_batch(self, input_batch):
        self.io_binding.bind_cpu_input(  # No-op if gpu is not available
            self.input_name,
            input_batch,
        )

        for name in self.output_names:
            self.io_binding.bind_output(name)

        # run session to get outputs
        if self._run_options is None:
            self.sess.run_with_iobinding(self.io_binding)
        else:
            self.sess.run_with_iobinding(self.io_binding, self._run_options)
        ort_outputs = self.io_binding.copy_outputs_to_cpu()  # No copy is performed if gpu is not available
        return ort_outputs



import json
from typing import List, Optional

import click
import yaml
from pydantic import NonNegativeFloat

from ..client import CaseType, HLJSONEncoder, TaskStatusEnum, TaskType, TrainingRunType
from ..core import GQLBaseModel


class UpdateTaskPayload(GQLBaseModel):
    errors: List[str]
    task: Optional[TaskType] = None


@click.group("task")
@click.pass_context
def task_group(ctx):
    pass


@task_group.command("read")
@click.option(
    "-i",
    "--ids",
    type=str,
    required=False,
    multiple=True,
)
@click.option(
    "-t",
    "--task-type",
    type=click.Choice(["TrainModel", "EvaluateAgent"], case_sensitive=True),
    required=False,
)
@click.option(
    "-r",
    "--training-run-id",
    type=int,
    required=False,
)
@click.pass_context
def read(ctx, ids, task_type, training_run_id):
    """Read task(s) by ID"""
    client = ctx.obj["client"]
    result = None

    if not ids and training_run_id is None:
        print("Error: Provide ids or training-run-id")
        return

    if task_type is None:
        result = []
        if len(ids) > 0:
            for id in ids:
                result.append(
                    client.task(
                        return_type=TaskType,
                        id=id,
                    ).gql_dict()
                )
    else:
        # TODO: Refactor once tasks are linked to training runs
        training_run = client.trainingRun(
            return_type=TrainingRunType,
            id=training_run_id,
        ).gql_dict()

        if "trainingConfig" in training_run:
            result = training_run["trainingConfig"]
        else:
            print("Error: no training config found in training run")

    if result is not None:
        click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("create")
@click.option(
    "-w",
    "--workflow-order-id",
    type=str,
    required=True,
    help="Project order ID",
)
@click.option("-f", "--file-ids", type=int, required=True, help="File IDs", multiple=True)
@click.pass_context
def create(ctx, workflow_order_id, file_ids):
    """Create task(s) in a process order"""
    client = ctx.obj["client"]

    class AddFilesToWorkflowOrderPayload(GQLBaseModel):
        errors: List[str]
        tasks: List[TaskType]

    result = client.addFilesToWorkflowOrder(
        return_type=AddFilesToWorkflowOrderPayload,
        workflowOrderId=workflow_order_id,
        fileIds=file_ids,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("update")
@click.option(
    "-i",
    "--id",
    type=str,
    required=True,
    help="The IDs of the task to update. Comma separate",
)
@click.option(
    "-n",
    "--name",
    type=str,
    required=False,
    help="The name of this task",
)
@click.option("-d", "--description", type=str, required=False, help="The description of this task")
@click.option(
    "-s",
    "--status",
    type=TaskStatusEnum,
    required=False,
    help="Status of task",
)
@click.option(
    "-t",
    "--tags",
    type=str,
    required=False,
    help="Tags for task",
    multiple=True,
)
@click.option(
    "-p",
    "--parameters",
    type=str,
    required=False,
    help="Aiko task parameters",
)
@click.option(
    "-r",
    "--requested-by-id",
    type=int,
    required=False,
    help="ID of the requester",
)
@click.option(
    "-l",
    "--leased-until",
    type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S%z"]),
    required=False,
    help="When to lease task(s) until",
)
@click.option(
    "-a",
    "--leased-by-agent-id",
    type=str,
    required=False,
    help="ID of the leasing agent",
)
@click.option(
    "-p",
    "--leased-by-pipeline-instance-id",
    type=str,
    required=False,
    help="ID of the leasing pipeline instance",
)
@click.pass_context
def update(
    ctx,
    id,
    name,
    description,
    status,
    tags,
    parameters,
    requested_by_id,
    leased_until,
    leased_by_agent_id,
    leased_by_pipeline_instance_id,
):
    client = ctx.obj["client"]

    if leased_until is not None:
        leased_until = leased_until.isoformat()

    kwargs = {
        k: v
        for k, v in dict(
            name=name,
            description=description,
            status=status,
            tags=tags,
            parameters=parameters,
            leasedUntil=leased_until,
            requestedById=requested_by_id,
            leasedByAgentId=leased_by_agent_id,
            leasedByPipelineInstanceId=leased_by_pipeline_instance_id,
        ).items()
        if v is not None
    }

    for i in id.split(","):
        i = i.strip()

        result = client.updateTask(return_type=UpdateTaskPayload, id=i, **kwargs).gql_dict()

        click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("lease")
@click.option(
    "-i",
    "--id",
    type=str,
    required=False,
    help="The ID of the task to lease",
)
@click.option(
    "-l",
    "--leased-until",
    type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S%z"]),
    required=True,
    help="When to lease task until",
)
@click.option(
    "-a",
    "--leased-by-agent-id",
    type=str,
    required=False,
    help="Agent ID to lease task by",
)
@click.option(
    "-p",
    "--leased-by-pipeline-instance-id",
    type=str,
    required=False,
    help="Pipeline instance ID to lease task by",
)
@click.pass_context
def lease(
    ctx,
    id,
    leased_until,
    leased_by_agent_id,
    leased_by_pipeline_instance_id,
):
    """Lease task"""
    client = ctx.obj["client"]

    if leased_by_agent_id is None and leased_by_pipeline_instance_id is None:
        raise ValueError("Error: One of leased-by-agent-id or leased-by-pipeline-instance-id must be set")

    result = client.updateTask(
        return_type=UpdateTaskPayload,
        id=id,
        leasedUntil=leased_until.isoformat(),
        leasedByAgentId=leased_by_agent_id,
        leasedByPipelineInstanceId=leased_by_pipeline_instance_id,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("re-lease")
@click.option(
    "-i",
    "--id",
    type=str,
    required=False,
    help="The ID of the task to re-lease",
)
@click.option(
    "-l",
    "--leased-until",
    type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S%z"]),
    required=True,
    help="When to lease task until",
)
@click.pass_context
def re_lease(
    ctx,
    id,
    leased_until,
):
    """Re-lease task"""
    client = ctx.obj["client"]

    result = client.updateTask(
        return_type=UpdateTaskPayload,
        id=id,
        leasedUntil=leased_until.isoformat(),
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("mark-with-status")
@click.option(
    "-i",
    "--id",
    type=str,
    required=False,
    help="The ID of the task to re-lease",
)
@click.option(
    "-s",
    "--status",
    type=str,
    required=True,
    help="The status to give the task",
)
@click.pass_context
def mark_with_status(
    ctx,
    id,
    status,
):
    """Mark with status"""
    client = ctx.obj["client"]

    result = client.updateTask(
        return_type=UpdateTaskPayload,
        id=id,
        status=status,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("lease-from-steps")
@click.option(
    "-l",
    "--leased-until",
    type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S%z"]),
    required=True,
    help="When to lease tasks until",
)
@click.option(
    "-a",
    "--leased-by-agent-id",
    type=str,
    required=False,
    help="Agent ID to lease tasks by",
)
@click.option(
    "-p",
    "--leased-by-pipeline-instance-id",
    type=str,
    required=False,
    help="Pipeline instance ID to lease tasks by",
)
@click.option(
    "-s",
    "--step-id",
    type=str,
    required=True,
    help="Lease tasks belonging to step(s)",
    multiple=True,
)
@click.option(
    "-c",
    "--count",
    type=int,
    required=True,
    help="Number of tasks to lease",
)
@click.pass_context
def lease_from_steps(
    ctx,
    leased_until,
    leased_by_agent_id,
    leased_by_pipeline_instance_id,
    step_id,
    count,
):
    """Lease task(s) belonging to step(s)"""
    client = ctx.obj["client"]

    class LeaseTaskPayload(GQLBaseModel):
        errors: List[str]
        tasks: Optional[List[TaskType]] = None

    result = client.leaseTasksFromSteps(
        return_type=LeaseTaskPayload,
        leasedUntil=leased_until.isoformat(),
        leasedByAgentId=leased_by_agent_id,
        leasedByPipelineInstanceId=leased_by_pipeline_instance_id,
        stepIds=step_id,
        count=count,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("unlease")
@click.option(
    "-i",
    "--id",
    type=str,
    required=True,
    help="The ID of the task to unlease",
)
@click.pass_context
def unlease(ctx, id):
    """Unlease task"""
    client = ctx.obj["client"]

    result = client.updateTask(
        return_type=UpdateTaskPayload,
        id=id,
        leasedUntil=None,
        leasedByAgentId=None,
        leasedByPipelineInstanceId=None,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))


@task_group.command("delete")
@click.option(
    "-i",
    "--id",
    type=str,
    required=True,
    help="The ID of the task to delete",
)
@click.pass_context
def delete(ctx, id):
    """Delete task"""
    client = ctx.obj["client"]

    class DeleteTaskPayload(GQLBaseModel):
        errors: List[str]
        task: Optional[TaskType] = None

    result = client.deleteTask(
        return_type=DeleteTaskPayload,
        id=id,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))



from collections import defaultdict

import numpy as np
import ultralytics


class WeightedSamplingMixin:
    """Shared weighted-sampling utilities for classification/detection/segmentation datasets."""

    def _init_weighted_sampling(self):

        self.train_mode = (
            "train" in str(getattr(self, "prefix", "")).lower()
        )  # standardise how we detect train mode (works for both datasets)

        self.counts = self.count_instances()
        self.class_weights = self.calculate_class_weights(self.counts)
        self.item_weights = self.calculate_item_weights(self.class_weights)
        self.probabilities = self.normalize_probs(self.item_weights)

    def calculate_class_weights(self, counts: np.ndarray) -> np.ndarray:
        total = float(np.sum(counts))
        w = total / counts.astype(np.float64)
        return w

    def normalize_probs(self, weights: np.ndarray) -> np.ndarray:
        s = float(np.sum(weights))
        if s <= 0 or not np.isfinite(s):
            print("Falling back to uniform sampling")
            return np.ones_like(weights, dtype=np.float64) / max(len(weights), 1)

        p = weights / s
        p = np.clip(p, 0.0, 1.0)  # numerical safety
        p = p / float(np.sum(p))  # renormalise after clipping
        return p.astype(np.float64)

    def weighted_index(self, n: int) -> int:
        return int(np.random.choice(n, p=self.probabilities))


class WeightedClassificationDataset(ultralytics.data.dataset.ClassificationDataset, WeightedSamplingMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_weighted_sampling()

    def count_instances(self):
        """
        Count the number of instances per class

        Returns:
            dict: A dict containing the counts for each class.
        """
        counts = defaultdict(int)
        for _, label, _, _ in self.samples:
            counts[label] += 1

        if counts:
            max_label = max(counts.keys())
        else:
            max_label = -1

        counts = np.array([counts[i] for i in range(max_label + 1)], dtype=np.float64)
        counts[counts == 0] = 1.0
        return counts

    def calculate_item_weights(self, class_weights: np.ndarray) -> np.ndarray:
        weights = []

        for sample in self.samples:
            label = int(sample[1])  # class label for this sample
            weight = class_weights[label]  # lookup weight for that class
            weights.append(weight)

        return np.array(weights, dtype=np.float64)

    def __getitem__(self, index):
        """
        Return transformed label information based on the sampled index.
        """
        # Don't use for validation
        if not self.train_mode:
            return super().__getitem__(index)
        else:
            index = self.weighted_index(len(self.samples))

            # print(f"INDEX: {index}")
            return super().__getitem__(index)


class WeightedClassificationTrainer(ultralytics.models.yolo.classify.train.ClassificationTrainer):

    def build_dataset(self, img_path: str, mode: str = "train", batch=None):
        if mode == "train":
            return WeightedClassificationDataset(root=img_path, args=self.args, augment=True, prefix=mode)
        else:
            return ultralytics.data.dataset.ClassificationDataset(
                root=img_path, args=self.args, augment=False, prefix=mode
            )


class WeightedDetectionDataset(ultralytics.data.dataset.YOLODataset, WeightedSamplingMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_weighted_sampling()

    def _get_image_classes(self, label_item) -> np.ndarray:

        if isinstance(label_item, dict) and "cls" in label_item:
            cls = label_item["cls"]
            try:
                cls = np.array(cls).astype(int).reshape(-1)
            except Exception:
                return np.array([], dtype=int)
            return cls
        return np.array([], dtype=int)

    def count_instances(self) -> np.ndarray:

        counts = defaultdict(int)

        for lb in self.labels:  # list[dict]
            cls = self._get_image_classes(lb)
            if cls.size == 0:
                continue
        nc = int(getattr(self, "nc", 0) or 0)
        arr = np.array([counts[i] for i in range(nc)], dtype=np.float64)

        arr[arr == 0] = 1.0  # avoid division by zero for missing classes
        return arr

    def calculate_item_weights(self, class_weights: np.ndarray) -> np.ndarray:
        n = len(self.labels)
        img_w = np.ones(n, dtype=np.float64)

        for i, lb in enumerate(self.labels):
            cls = self._get_image_classes(lb)
            if cls.size == 0:
                img_w[i] = 1.0
                continue

            cls = cls[(cls >= 0) & (cls < len(class_weights))]
            if cls.size == 0:
                img_w[i] = 1.0
                continue

            img_w[i] = float(np.mean(class_weights[cls]))

        return img_w

    def __getitem__(self, index):
        if not self.train_mode:
            return super().__getitem__(index)

        index = self.weighted_index(len(self.labels))
        # print(f"INDEX: {index}")  # debug
        return super().__getitem__(index)


class WeightedDetectionTrainer(ultralytics.models.yolo.detect.train.DetectionTrainer):

    def build_dataset(self, img_path: str, mode: str = "train", batch=None):
        if mode == "train":
            return WeightedDetectionDataset(
                img_path=img_path,
                imgsz=self.args.imgsz,
                batch_size=batch,
                augment=True,
                hyp=self.args,
                rect=False,
                stride=32,
                pad=0.0,
                single_cls=self.args.single_cls,
                data=self.data,
                prefix=mode,
            )
        else:
            return ultralytics.data.dataset.YOLODataset(
                img_path=img_path,
                imgsz=self.args.imgsz,
                batch_size=batch,
                augment=False,
                hyp=self.args,
                rect=True,
                stride=32,
                pad=0.5,
                single_cls=self.args.single_cls,
                data=self.data,
                prefix=mode,
            )



"""Configuration classes for LLM capability."""

from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field, model_validator

from highlighter.agent.capabilities.base_capability import Capability

__all__ = [
    "LLMProfile",
    "VisionConfig",
    "MonitoringConfig",
    "OutputConfig",
    "LLMInitParameters",
    "LLMStreamParameters",
]


class LLMProfile(str, Enum):
    """Pre-configured profiles for common scenarios"""

    SIMPLE = "simple"  # Text-only processing
    VISION = "vision"  # Vision + observations
    STRUCTURED = "structured"  # JSON/Entity output
    CONVERSATION = "conversation"  # Multi-turn dialogue


class VisionConfig(BaseModel):
    """Vision processing settings"""

    enabled: bool = True
    max_images_per_request: int = 5
    resize_images: bool = True
    max_image_dimension: int = 1568
    image_quality: int = 85
    preserve_format: bool = True  # PNG for alpha channel


class MonitoringConfig(BaseModel):
    """Cost tracking and rate limiting"""

    track_metrics: bool = True
    max_requests_per_minute: int = 60
    max_cost_per_stream: Optional[float] = None


class OutputConfig(BaseModel):
    """Output format configuration"""

    format: str = "text"  # 'text', 'json', 'entities'
    create_entities: bool = False
    schema_version: str = "v1"


class LLMInitParameters(Capability.InitParameters):
    """Initialization parameters for LLM capability"""

    # Core configuration (3-5 parameters)
    profile: LLMProfile = LLMProfile.SIMPLE
    model: str = "claude-3-5-sonnet-20241022"
    prompt_template: Optional[str] = None
    system_prompt_template: Optional[str] = None

    # Frame windowing and taxonomy
    frame_window_size: Optional[int] = None
    time_window_seconds: Optional[float] = None
    taxonomy: Optional[dict] = None  # Will be converted to TaxonomyConfig
    content_filters: Optional[list] = None  # Will be converted to List[ContentFilterConfig]

    # Optional advanced configuration
    vision: Optional[VisionConfig] = None
    monitoring: Optional[MonitoringConfig] = None
    output: Optional[OutputConfig] = None

    # Advanced options
    use_jinja2: bool = True
    validate_template: bool = True
    completion_kwargs: dict = Field(default_factory=dict)
    mock_response: bool = False
    num_retries: int = 8

    # Legacy compatibility
    cache_system_prompt: bool = False

    @model_validator(mode="after")
    def apply_profile_defaults(self) -> "LLMInitParameters":
        """Apply profile-specific defaults"""
        if self.profile == LLMProfile.SIMPLE:
            if self.vision is None:
                self.vision = VisionConfig(enabled=False)
        elif self.profile == LLMProfile.VISION:
            if self.vision is None:
                self.vision = VisionConfig(enabled=True)
            if self.monitoring is None:
                self.monitoring = MonitoringConfig(track_metrics=True)
        elif self.profile == LLMProfile.STRUCTURED:
            if self.output is None:
                self.output = OutputConfig(format="json")
        elif self.profile == LLMProfile.CONVERSATION:
            if self.monitoring is None:
                self.monitoring = MonitoringConfig(track_metrics=True)

        # Set defaults for None configs
        self.vision = self.vision or VisionConfig()
        self.monitoring = self.monitoring or MonitoringConfig()
        self.output = self.output or OutputConfig()
        return self

    @model_validator(mode="after")
    def validate_config_consistency(self) -> "LLMInitParameters":
        """Validate configuration consistency"""
        if self.vision.enabled and not self._model_supports_vision(self.model):
            raise ValueError(
                f"Model {self.model} does not support vision. "
                "Set vision.enabled=False or use a vision model."
            )

        if self.output.format == "entities":
            self.output.create_entities = True

        # Ensure at least one window size is set, default to frame_window_size=1
        if self.frame_window_size is None and self.time_window_seconds is None:
            self.frame_window_size = 1

        # Validate window sizes if set
        if self.frame_window_size is not None and self.frame_window_size < 1:
            raise ValueError("frame_window_size must be at least 1")
        if self.time_window_seconds is not None and self.time_window_seconds <= 0:
            raise ValueError("time_window_seconds must be greater than 0")

        return self

    @staticmethod
    def _model_supports_vision(model: str) -> bool:
        """Check if model supports vision"""
        vision_models = [
            "claude-3",
            "gpt-4o",
            "gpt-4-turbo",
            "gemini-1.5",
            "gemini-pro-vision",
        ]
        return any(vm in model.lower() for vm in vision_models)


class LLMStreamParameters(LLMInitParameters):
    """Per-stream parameters"""

    prompt_template_path: Optional[str] = None
    system_prompt_template_path: Optional[str] = None
    template_vars: dict = Field(default_factory=dict)

    # Conversation config (state in stream.variables)
    enable_conversation: bool = False
    max_history_turns: int = 10

    # Legacy compatibility
    strategy: Optional[str] = None



import logging
import time
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from uuid import UUID

from pydantic import Field

from highlighter.agent.capabilities._crop_processor import CropProcessor
from highlighter.agent.capabilities.base_capability import Capability, StreamEvent
from highlighter.client.base_models import Entity
from highlighter.client.base_models.annotation import Annotation
from highlighter.client.base_models.entities import Entities
from highlighter.client.gql_client import HLClient
from highlighter.client.training_runs import TrainingRunArtefactType
from highlighter.core.data_models import DataSample
from highlighter.predictors.onnx_yolov8 import OnnxYoloV8
from highlighter.predictors.output_category_descriptor import OutputCategoryDescriptor
from highlighter.predictors.prediction import Prediction

__all__ = ["YoloBoxClassifier"]

logger = logging.getLogger(__name__)


class YoloBoxClassifier(Capability):
    """
    A box classifier capability that uses YOLO to classify existing bounding boxes.

    This capability takes entities with bounding box or polygon annotations from upstream
    (e.g., from an object detector) and classifies each box using a YOLO classifier model.
    The classification results are added as observations to the existing annotations.
    """

    class InitParameters(Capability.InitParameters):
        onnx_file: Optional[str] = None
        training_run_artefact_id: Optional[UUID] = None
        num_classes: int
        class_lookup: Optional[Dict[int | str, OutputCategoryDescriptor]] = None
        conf_thresh: float = 0.1
        nms_iou_thresh: float = 0.5
        is_absolute: bool = True
        crop_processor: CropProcessor = Field(..., default_factory=lambda: CropProcessor())

    def __init__(self, context):
        super().__init__(context)

        onnx_file = self.init_parameters.onnx_file
        if onnx_file is None:
            training_run_artefact_id = self.init_parameters.training_run_artefact_id
            if training_run_artefact_id is None:
                raise ValueError("Must provide `onnx_file` or `training_run_artefact_id`")
            training_run_artefact = HLClient.get_client().trainingRunArtefact(
                return_type=TrainingRunArtefactType, id=training_run_artefact_id
            )
            onnx_file = training_run_artefact.file_url

        artefact_cache_dir = Path.home() / ".cache" / "highlighter" / "agents" / "artefacts"
        self._predictor = OnnxYoloV8(
            onnx_file,
            self.init_parameters.num_classes,
            class_lookup=self.init_parameters.class_lookup,
            conf_thresh=self.init_parameters.conf_thresh,
            nms_iou_thresh=self.init_parameters.nms_iou_thresh,
            is_absolute=self.init_parameters.is_absolute,
            artefact_cache_dir=artefact_cache_dir,
        )

        self._crop_processor = self.init_parameters.crop_processor

    def process_frame(
        self, stream, data_samples: List[DataSample], entities: Dict[UUID, Entity]
    ) -> Tuple[StreamEvent, Dict]:
        start = time.perf_counter()

        # ToDo: entities should always be an `Entities` object
        if isinstance(entities, dict):
            entities = Entities(entities)

        crops = self._crop_processor.crop_annotations(entities.all_annotations())
        preds = self._predictor.predict(crops)
        for predictions_from_annotation, annotation in zip(preds, entities.all_annotations()):
            for prediction in predictions_from_annotation:
                # convert to Observation and add to the source Annotation
                prediction.add_to_annotation(annotation)

        end = time.perf_counter()
        self.logger.debug(f"YoloBoxClassifier processed in {end - start:.6f} seconds")

        return StreamEvent.OKAY, {"entities": entities}



from typing import Any, List
from uuid import UUID

from pydantic import BaseModel, Field, model_validator

from highlighter.core.const import OBJECT_CLASS_ATTRIBUTE_UUID


class ObjectClassDescriptor(BaseModel):
    """
    ObjectClassDescriptor([object_class_uuid, object_class_label])
    """

    uuid: UUID
    label: str

    @model_validator(mode="before")
    def validate(cls, v):
        if isinstance(v, dict):
            return v
        elif len(v) == 2:
            return dict(zip(["uuid", "label"], v))
        else:
            raise ValueError(f"Unkonwn {cls.__name__} args: {v}")


class ObjectClassDescriptorWithConfidenceThreshold(BaseModel):
    """
    ObjectClassDescriptorWithConfidenceThreshold(
        [object_class_uuid, object_class_label, confidence_threshold])
    """

    uuid: UUID
    label: str
    confidence_threshold: float = Field(ge=0, le=1)

    @model_validator(mode="before")
    def validate(cls, v):
        if isinstance(v, dict):
            return v
        elif len(v) == 3:
            return dict(zip(["uuid", "label", "confidence_threshold"], v))
        else:
            raise ValueError(f"Unkonwn {cls.__name__} args: {v}")


class EnumDescriptor(BaseModel):
    """
    EnumDescriptor([attribute_id, attribute_label, enum_id, enum_label])
    """

    attribute_id: UUID
    attribute_label: str
    enum_id: UUID
    enum_label: str

    @model_validator(mode="before")
    def validate(cls, v):
        if isinstance(v, dict):
            return v
        elif len(v) == 4:
            return dict(zip(["attribute_id", "attribute_label", "enum_id", "enum_label"], v))
        else:
            raise ValueError(f"Unkonwn {cls.__name__} args: {v}")


class EnumDescriptorWithConfidenceThreshold(BaseModel):
    """
    EnumDescriptorWithConfidenceThreshold(
        [attribute_id, attribute_label, enum_id, enum_label, confidence_threshold])
    """

    attribute_id: UUID
    attribute_label: str
    enum_id: UUID
    enum_label: str
    confidence_threshold: float = Field(ge=0, le=1)

    @model_validator(mode="before")
    def validate(cls, v):
        if isinstance(v, dict):
            return v
        elif len(v) == 5:
            return dict(
                zip(["attribute_id", "attribute_label", "enum_id", "enum_label", "confidence_threshold"], v)
            )
        else:
            raise ValueError(f"Unkonwn {cls.__name__} args: {v}")

    @classmethod
    def from_output_category_descriptor(
        cls, output_category_descriptor: "OutputCategoryDescriptor", default_confidence_threshold: float = 0.0
    ):
        if isinstance(output_category_descriptor, ObjectClassDescriptor):
            return cls(
                attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                attribute_label="object_class",
                enum_id=output_category_descriptor.uuid,
                enum_label=output_category_descriptor.label,
                confidence_threshold=default_confidence_threshold,
            )
        if isinstance(output_category_descriptor, ObjectClassDescriptorWithConfidenceThreshold):
            return cls(
                attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                attribute_label="object_class",
                enum_id=output_category_descriptor.uuid,
                enum_label=output_category_descriptor.label,
                confidence_threshold=output_category_descriptor.confidence_threshold,
            )
        if isinstance(output_category_descriptor, EnumDescriptor):
            return cls(
                **output_category_descriptor.model_dump(), confidence_threshold=default_confidence_threshold
            )
        if isinstance(output_category_descriptor, EnumDescriptorWithConfidenceThreshold):
            return output_category_descriptor


OutputCategoryDescriptor = (
    ObjectClassDescriptor
    | ObjectClassDescriptorWithConfidenceThreshold
    | EnumDescriptor
    | EnumDescriptorWithConfidenceThreshold
)



import json
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Dict

from click.testing import CliRunner

from highlighter.cli.highlighter_cli import highlighter_group

__all__ = ["HLAgentCliRunner"]


class HLAgentCliRunner:

    def __init__(self, definition: Dict):
        self._tmpdir = TemporaryDirectory()

        self.definition = definition
        self.definition_path = f"{self._tmpdir.name}/definition.json"
        with open(self.definition_path, "w") as f:
            json.dump(self.definition, f)

        self.queue_response = Queue()

    def hl_agent_start(self, opts: str = "", args: str = "", click_invoke_input=None, **click_runner_kwargs):
        runner = CliRunner(**click_runner_kwargs)
        cli_cmd = f"agent start {opts} {self.definition_path} {args}"
        result = runner.invoke(
            highlighter_group,
            cli_cmd.split(),
            obj={"queue_response": self.queue_response},
            input=click_invoke_input,
            catch_exceptions=False,
        )
        return result



"""
Prepend Gaussian blur and resize preprocessing to NeuFlow ONNX model.

This creates a single ONNX model that:
1. Takes raw uint8 HWC images as input
2. Applies Gaussian blur (kernel provided at runtime)
3. Resizes to model input dimensions
4. Runs optical flow estimation
5. Resizes flow output back to original dimensions

Usage:
    python -m highlighter.agent.capabilities.image_to_scalar._neuflow_preprocessing \
        input_model.onnx output_model.onnx
"""

import numpy as np
import onnx
from onnx import TensorProto, helper, numpy_helper

MAX_KERNEL_SIZE = 7


def make_gaussian_kernel(kernel_size: int, sigma: float) -> np.ndarray:
    """Create a 2D Gaussian kernel for depthwise convolution."""
    coords = np.arange(kernel_size) - kernel_size // 2
    g = np.exp(-(coords**2) / (2 * sigma**2))
    kernel_2d = np.outer(g, g)
    kernel_2d = kernel_2d / kernel_2d.sum()
    # Shape for depthwise conv: (out_channels=3, in_channels/groups=1, H, W)
    kernel_4d = np.tile(kernel_2d[np.newaxis, np.newaxis, :, :], (3, 1, 1, 1))
    # Zero-pad to max kernel size
    if kernel_size < MAX_KERNEL_SIZE:
        pad = (MAX_KERNEL_SIZE - kernel_size) // 2
        kernel_4d = np.pad(kernel_4d, ((0, 0), (0, 0), (pad, pad), (pad, pad)))
    return kernel_4d.astype(np.float32)


def make_identity_kernel() -> np.ndarray:
    """Create identity kernel (no-op convolution)."""
    kernel = np.zeros((MAX_KERNEL_SIZE, MAX_KERNEL_SIZE), dtype=np.float32)
    kernel[MAX_KERNEL_SIZE // 2, MAX_KERNEL_SIZE // 2] = 1.0
    return np.tile(kernel[np.newaxis, np.newaxis, :, :], (3, 1, 1, 1))


def _make_preprocess_nodes(
    input_name: str,
    output_name: str,
    model_h: int,
    model_w: int,
    prefix: str,
) -> list:
    """Create preprocessing nodes for a single image input."""
    nodes = []
    pad = MAX_KERNEL_SIZE // 2

    # HWC -> CHW
    nodes.append(
        helper.make_node(
            "Transpose",
            inputs=[input_name],
            outputs=[f"{prefix}chw"],
            perm=[2, 0, 1],
            name=f"{prefix}transpose",
        )
    )

    # Add batch dim: CHW -> BCHW
    nodes.append(
        helper.make_node(
            "Unsqueeze",
            inputs=[f"{prefix}chw", "unsqueeze_axes"],
            outputs=[f"{prefix}bchw"],
            name=f"{prefix}unsqueeze",
        )
    )

    # Cast uint8 -> float32
    nodes.append(
        helper.make_node(
            "Cast",
            inputs=[f"{prefix}bchw"],
            outputs=[f"{prefix}float"],
            to=TensorProto.FLOAT,
            name=f"{prefix}cast",
        )
    )

    # Normalize [0, 255] -> [0, 1]
    nodes.append(
        helper.make_node(
            "Div",
            inputs=[f"{prefix}float", "scale_255"],
            outputs=[f"{prefix}normalized"],
            name=f"{prefix}normalize",
        )
    )

    # Gaussian blur via depthwise convolution (kernel is a runtime input)
    nodes.append(
        helper.make_node(
            "Conv",
            inputs=[f"{prefix}normalized", "blur_kernel"],
            outputs=[f"{prefix}blurred"],
            kernel_shape=[MAX_KERNEL_SIZE, MAX_KERNEL_SIZE],
            pads=[pad, pad, pad, pad],
            group=3,
            name=f"{prefix}blur",
        )
    )

    # Resize to model input dimensions
    nodes.append(
        helper.make_node(
            "Resize",
            inputs=[f"{prefix}blurred", "", "", "resize_shape"],
            outputs=[output_name],
            mode="linear",
            coordinate_transformation_mode="asymmetric",
            name=f"{prefix}resize",
        )
    )

    return nodes


def prepend_preprocessing_to_neuflow(
    neuflow_path: str,
    output_path: str,
) -> None:
    """
    Modify NeuFlow ONNX model to include preprocessing.

    New model signature:
        Inputs:
            - prev_image: (H, W, 3) uint8
            - curr_image: (H, W, 3) uint8
            - blur_kernel: (3, 1, K, K) float32 - Gaussian or identity kernel
        Outputs:
            - flow: (H, W, 2) float32
    """
    # Load original model
    original_model = onnx.load(neuflow_path)
    original_graph = original_model.graph

    # Get model input dimensions from original model
    input_shape = original_graph.input[0].type.tensor_type.shape
    model_h = input_shape.dim[2].dim_value
    model_w = input_shape.dim[3].dim_value
    print(f"NeuFlow model input size: {model_h}x{model_w}")

    # Build new nodes list
    new_nodes = []
    new_initializers = list(original_graph.initializer)

    # --- Preprocessing for prev_image ---
    new_nodes.extend(
        _make_preprocess_nodes(
            input_name="prev_image",
            output_name="prev_processed",
            model_h=model_h,
            model_w=model_w,
            prefix="prev_",
        )
    )

    # --- Preprocessing for curr_image ---
    new_nodes.extend(
        _make_preprocess_nodes(
            input_name="curr_image",
            output_name="curr_processed",
            model_h=model_h,
            model_w=model_w,
            prefix="curr_",
        )
    )

    # Add shared initializers (constants used by preprocessing)
    new_initializers.extend(
        [
            numpy_helper.from_array(np.array([0], dtype=np.int64), name="unsqueeze_axes"),
            numpy_helper.from_array(np.array([0], dtype=np.int64), name="squeeze_axes"),
            numpy_helper.from_array(np.array([255.0], dtype=np.float32), name="scale_255"),
            numpy_helper.from_array(np.array([1, 3, model_h, model_w], dtype=np.int64), name="resize_shape"),
        ]
    )

    # --- Rename original model inputs to connect to preprocessing outputs ---
    original_input_names = [inp.name for inp in original_graph.input]
    print(f"Original input names: {original_input_names}")

    input_mapping = {
        original_input_names[0]: "prev_processed",
        original_input_names[1]: "curr_processed",
    }

    # Copy and rename original model nodes
    for node in original_graph.node:
        new_node = onnx.NodeProto()
        new_node.CopyFrom(node)

        # Rename inputs if they match original model inputs
        new_inputs = []
        for inp in node.input:
            new_inputs.append(input_mapping.get(inp, inp))
        new_node.ClearField("input")
        new_node.input.extend(new_inputs)

        new_nodes.append(new_node)

    # --- Postprocessing: resize flow back to original dimensions ---
    original_output_name = original_graph.output[0].name

    # Get original image dimensions for resizing output
    new_nodes.append(
        helper.make_node(
            "Shape",
            inputs=["prev_image"],
            outputs=["input_shape"],
            name="get_input_shape",
        )
    )

    # Extract H, W from shape (H, W, 3)
    new_nodes.append(
        helper.make_node(
            "Slice",
            inputs=["input_shape", "slice_start", "slice_end"],
            outputs=["hw_shape"],
            name="extract_hw",
        )
    )
    new_initializers.append(numpy_helper.from_array(np.array([0], dtype=np.int64), name="slice_start"))
    new_initializers.append(numpy_helper.from_array(np.array([2], dtype=np.int64), name="slice_end"))

    # Build output shape: [1, 2, H, W]
    new_nodes.append(
        helper.make_node(
            "Concat",
            inputs=["batch_channels", "hw_shape"],
            outputs=["output_resize_shape"],
            axis=0,
            name="build_output_shape",
        )
    )
    new_initializers.append(numpy_helper.from_array(np.array([1, 2], dtype=np.int64), name="batch_channels"))

    # Resize flow to original dimensions
    new_nodes.append(
        helper.make_node(
            "Resize",
            inputs=[original_output_name, "", "", "output_resize_shape"],
            outputs=["flow_resized"],
            mode="linear",
            coordinate_transformation_mode="asymmetric",
            name="resize_flow_output",
        )
    )

    # Squeeze batch dim: (1, 2, H, W) -> (2, H, W)
    new_nodes.append(
        helper.make_node(
            "Squeeze",
            inputs=["flow_resized", "squeeze_axes"],
            outputs=["flow_squeezed"],
            name="squeeze_flow",
        )
    )

    # Transpose: (2, H, W) -> (H, W, 2)
    new_nodes.append(
        helper.make_node(
            "Transpose",
            inputs=["flow_squeezed"],
            outputs=["flow_output"],
            perm=[1, 2, 0],
            name="transpose_flow",
        )
    )

    # --- Create new graph ---
    new_inputs = [
        helper.make_tensor_value_info("prev_image", TensorProto.UINT8, ["H", "W", 3]),
        helper.make_tensor_value_info("curr_image", TensorProto.UINT8, ["H", "W", 3]),
        helper.make_tensor_value_info(
            "blur_kernel", TensorProto.FLOAT, [3, 1, MAX_KERNEL_SIZE, MAX_KERNEL_SIZE]
        ),
    ]

    new_outputs = [
        helper.make_tensor_value_info("flow_output", TensorProto.FLOAT, ["H", "W", 2]),
    ]

    new_graph = helper.make_graph(
        nodes=new_nodes,
        name="neuflow_with_preprocessing",
        inputs=new_inputs,
        outputs=new_outputs,
        initializer=new_initializers,
    )

    # Create new model
    new_model = helper.make_model(new_graph, opset_imports=original_model.opset_import)
    new_model.ir_version = original_model.ir_version

    # Validate and save
    onnx.checker.check_model(new_model)
    onnx.save(new_model, output_path)
    print(f"Saved modified model to: {output_path}")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Add preprocessing to NeuFlow model")
    parser.add_argument("input_model", help="Path to original NeuFlow ONNX model")
    parser.add_argument("output_model", help="Path for output model with preprocessing")

    args = parser.parse_args()

    prepend_preprocessing_to_neuflow(args.input_model, args.output_model)



import os
import uuid
from pathlib import Path
from typing import ClassVar, Optional
from uuid import UUID

from sqlalchemy import Column, String, event
from sqlmodel import Field, Relationship, SQLModel

from highlighter.client.gql_client import HLClient
from highlighter.core.config import HighlighterRuntimeConfig
from highlighter.core.data_models.account_mixin import AccountMixin
from highlighter.core.utilities import get_slug


# TODO Decide best way to get a session across SDK codebase
class DataFileSource(SQLModel, AccountMixin, table=True):
    data_dir: ClassVar[Optional[Path]] = None

    id: str = Field(default_factory=lambda: str(uuid.uuid4()), sa_column=Column(String, primary_key=True))
    url: str = Field(sa_column=Column(String, default=None, primary_key=True))
    payload: str = Field(sa_column=Column(String, default=None, primary_key=True))
    request_hash: str = Field(sa_column=Column(String, default=None, primary_key=True))
    data_file_id: Optional[UUID] = Field(default=None, foreign_key="datafile.file_id")
    data_file: Optional["DataFile"] = Relationship(back_populates="data_file_sources")

    @classmethod
    def get_data_dir(cls) -> Path:
        if cls.data_dir is None:
            # FIXME: How do we best pull HighlighterRuntimeConfig out of here?
            hl_data_models_dir = HighlighterRuntimeConfig.load().data_models_dir(
                HLClient.get_client().account_name
            )
            cls.data_dir = hl_data_models_dir / get_slug(cls.__qualname__)
        return cls.data_dir

    def get_response(self):
        if self.data_file is None:
            raise ValueError(
                f"Error when getting response for DataFileSource {self.id}: No associated data file found"
            )

        return self.data_file.content

    @property
    def path_to_content_file(self):
        if self.data_file is None:
            raise ValueError(
                f"Error when getting path to content file for DataFileSource {self.id}: No associated data file found"
            )

        return self.data_file.path_to_content_file

    def write_content_to_disk(self):
        if self.data_file is None:
            raise ValueError(
                f"Error when getting path to content file for DataFileSource {self.id}: No associated data file found"
            )

        self.data_file.write_content_to_disk()


def after_load(target, _):
    """
    Called when an object is loaded from the database
    """
    if not os.path.exists(target.data_file.path_to_content_file):
        raise ValueError(f"Error: file on disk not found when loading DataFileSource id {target.id}")


event.listen(DataFileSource, "load", after_load)



import logging
from datetime import datetime
from typing import Dict, NamedTuple, Optional, Set, Union
from uuid import UUID, uuid4

import celpy
from pydantic import BaseModel, Field, field_validator

from highlighter.core.enums import ContentTypeEnum
from highlighter.core.labeled_uuid import LabeledUUID

# ToDo: Consolidate this work with the data layer
# ToDo: Consolidate this work with the Entities object and Agents
# ToDo: Consolidate this work with Datasets

__all__ = ["ObservationsTable"]

logger = logging.getLogger(__name__)


class DotDict(dict):
    """
    Dictionary with dot notation access and recursive conversion.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        if key in self:
            del self[key]
        else:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __repr__(self):
        return f"Attribtues({dict.__repr__(self)})"


class ObservationsTable:
    class Row(BaseModel):

        model_config = {"arbitrary_types_allowed": True}

        class Stream(BaseModel):
            id: str

        class Entity(BaseModel):
            id: UUID

        class DataSample(BaseModel):
            recorded_at: datetime
            content_type: ContentTypeEnum
            stream_frame_index: int
            media_frame_index: int
            width: Optional[int] = None
            height: Optional[int] = None

        class Annotation(BaseModel):

            class Location(BaseModel):
                wkt: str
                xmin: int
                ymin: int
                xmax: int
                ymax: int

                @property
                def area(self):
                    return (self.xmax - self.xmin) * (self.ymax - self.ymin)

            id: UUID
            location: Optional[Location] = None

        class Attribute(BaseModel):

            class Category(BaseModel):
                id: UUID
                name: str

            value: Union[int, float, str, Category]
            occurred_at: datetime
            confidence: Optional[float] = None

            @field_validator("value", mode="before")
            def handel_labeled_uuid(cls, v):
                if isinstance(v, LabeledUUID):
                    return cls.Category(id=v, name=v.label)
                if isinstance(v, UUID):
                    return cls.Category(id=v, name="__UNDEFINED__")
                return v

        # class _NullAttribute(dict):
        #    """A safe null object that returns None for any key access"""

        #    def __getitem__(self, key):
        #        return None

        #    def __getattr__(self, key):
        #        return None

        #    def get(self, key, default=None):
        #        return default

        # class SafeAttributes(dict):

        #    def __init__(self, attributes):
        #        # Convert Pydantic models to dicts for CEL
        #        self._attributes = {}
        #        for k, v in attributes.items():
        #            attr = ObservationsTable.Row.Attribute(**v)
        #            # Store as dict for CEL field selection support
        #            self._attributes[k] = {
        #                "value": attr.value,
        #                "occurred_at": attr.occurred_at,
        #                "confidence": attr.confidence,
        #            }
        #        # Initialize dict with the attributes
        #        super().__init__(self._attributes)

        #    def __getattr__(self, k):
        #        val = self._attributes.get(k, None)
        #        if val and isinstance(val, dict):
        #            # Convert back to Attribute for Python attribute access
        #            return ObservationsTable.Row.Attribute(**val)
        #        return val

        #    def __getitem__(self, k):
        #        """Support dict-like access for CEL, returns safe null object for missing keys"""
        #        val = self._attributes.get(k, None)
        #        if val is None:
        #            return ObservationsTable.Row._NullAttribute()
        #        return val

        #    def get(self, k, default=None):
        #        """Support dict.get() for CEL"""
        #        return self._attributes.get(k, default)

        entity: Optional[Entity]
        stream: Stream
        data_sample: DataSample
        annotation: Optional[Annotation]
        attribute: DotDict
        id: UUID = Field(default_factory=uuid4)

        @field_validator("attribute", mode="before")
        @classmethod
        def convert_attribute_dict(cls, v):
            if isinstance(v, dict) or isinstance(v, DotDict):
                return DotDict(**{key: cls.Attribute(**value) for key, value in v.items()})
            return v

        def to_cel_ctx_dict(self, all_attr_keys: Set[str]):
            attributes = {k: {"value": None, "occurred_at": None, "confidence": None} for k in all_attr_keys}
            for k, v in self.attribute.items():
                attributes[k] = v.model_dump(mode="json")

            return {
                "id": str(self.id),
                "entity": self.entity.model_dump(mode="json") if self.entity else None,
                "stream": self.stream.model_dump(mode="json"),
                "data_sample": self.data_sample.model_dump(mode="json"),
                "annotation": self.annotation.model_dump(mode="json") if self.annotation else None,
                "attribute": attributes,
            }

    def __init__(self, rows: Dict[str, Row] = {}):
        self._rows = rows

    def __getitem__(self, key):
        """Support indexing for CEL"""
        return self._rows[key]

    def __iter__(self):
        """Support iteration for CEL"""
        return iter(self._rows)

    def __len__(self):
        """Support len() for CEL"""
        return len(self._rows)

    def clear(self):
        """Remove all rows once they've been consumed."""
        self._rows.clear()

    @classmethod
    def from_row_records(cls, records):

        rows = {}
        for r in records:
            row = ObservationsTable.Row(**r)
            rows[str(row.id)] = row
        return cls(rows)

    def get_cel_ctx_rows(self):
        # Collect all attribute keys across all rows
        all_attr_keys = set()
        for r in self._rows.values():
            all_attr_keys.update(r.attribute.keys())

        for r in self._rows.values():
            yield celpy.json_to_cel(r.to_cel_ctx_dict(all_attr_keys))

    def filter(self, expr):
        env = celpy.Environment()
        ast = env.compile(expr)
        program = env.program(ast)

        # Evaluate each row individually and catch errors
        result = []
        for row_ctx in self.get_cel_ctx_rows():
            try:
                # Evaluate expression with row fields directly in context
                matches = program.evaluate(row_ctx)
                if matches:
                    result.append(self._rows[str(row_ctx["id"])])
            except (celpy.evaluation.CELEvalError, TypeError):
                # If evaluation fails (e.g., None comparison), treat as non-matching
                pass

        return result

    def any(self, expr):
        """
        Check if any row matches the given CEL expression.

        Args:
            expr: A CEL expression to evaluate against each row

        Returns:
            bool: True if at least one row matches, False otherwise
        """
        env = celpy.Environment()
        ast = env.compile(expr)
        program = env.program(ast)

        for row_ctx in self.get_cel_ctx_rows():
            try:
                matches = program.evaluate(row_ctx)
                if matches:
                    return True
            except (celpy.evaluation.CELEvalError, TypeError):
                # If evaluation fails, treat as non-matching
                pass

        return False

    def all(self, expr):
        """
        Check if all rows match the given CEL expression.

        Args:
            expr: A CEL expression to evaluate against each row

        Returns:
            bool: True if all rows match, False otherwise
        """
        env = celpy.Environment()
        ast = env.compile(expr)
        program = env.program(ast)

        for row_ctx in self.get_cel_ctx_rows():
            try:
                matches = program.evaluate(row_ctx)
                if not matches:
                    return False
            except (celpy.evaluation.CELEvalError, TypeError):
                # If evaluation fails, treat as non-matching
                return False

        return True

    def show(self, log_extra=None):
        """Log a tabular representation of the rows"""
        if not self._rows:
            logger.info("Empty Observations Table", extra=log_extra)
            return

        # Collect all attribute keys across all rows
        all_attr_keys = sorted(set(key for row in self._rows.values() for key in row.attribute.keys()))

        # Define columns: id, entity_id, stream_id, recorded_at, then attributes
        columns = ["id", "entity_id", "stream_id", "recorded_at"] + all_attr_keys

        # Calculate column widths
        col_widths = {col: len(col) for col in columns}

        def get_attr_value_str(attr) -> str:
            attr_val = attr["value"] if isinstance(attr, dict) else attr.value

            if isinstance(attr_val, float):
                value_str = f"{attr_val:.3f}"
            else:
                value_str = str(attr_val)
            return value_str

        # Update widths based on data
        for row in self._rows.values():
            col_widths["id"] = max(col_widths["id"], len(str(row.id)[:8]))
            entity_id_str = str(row.entity.id)[:8] if row.entity else "None"
            col_widths["entity_id"] = max(col_widths["entity_id"], len(entity_id_str))
            col_widths["stream_id"] = max(col_widths["stream_id"], len(row.stream.id))
            col_widths["recorded_at"] = max(
                col_widths["recorded_at"], len(row.data_sample.recorded_at.strftime("%Y-%m-%d %H:%M"))
            )

            for attr_key in all_attr_keys:
                attr = row.attribute.get(attr_key)
                if attr:
                    value_str = get_attr_value_str(attr)

                    col_widths[attr_key] = max(col_widths[attr_key], len(value_str))

        # Build header
        header = " | ".join(col.ljust(col_widths[col]) for col in columns)
        separator = "-" * len(header)

        # Build all rows
        rows_output = []
        for row in self._rows.values():
            row_data = []
            row_data.append(str(row.id)[:8].ljust(col_widths["id"]))
            entity_id_str = str(row.entity.id)[:8] if row.entity else "None"
            row_data.append(entity_id_str.ljust(col_widths["entity_id"]))
            row_data.append(row.stream.id.ljust(col_widths["stream_id"]))
            row_data.append(
                row.data_sample.recorded_at.strftime("%Y-%m-%d %H:%M").ljust(col_widths["recorded_at"])
            )

            for attr_key in all_attr_keys:
                attr = row.attribute.get(attr_key)
                if attr:
                    value_str = get_attr_value_str(attr)
                    row_data.append(value_str.ljust(col_widths[attr_key]))
                else:
                    row_data.append("".ljust(col_widths[attr_key]))

            rows_output.append(" | ".join(row_data))

        # Log the entire table as a single message
        table_output = "\n".join([header, separator] + rows_output[-5:])
        logger.info(f"Observations Table (tail 5 of {len(self._rows)}):\n{table_output}", extra=log_extra)

    def add_entity(self, entity, data_sample, stream_id):
        new_rows = [
            ObservationsTable.Row(**r)
            for r in ObservationsTable.row_data_from_entity(entity, data_sample, stream_id)
        ]
        self._rows.update({str(r.id): r for r in new_rows})

    @staticmethod
    def row_data_from_entity(entity, data_sample, stream_id):
        rows = []
        # Process each annotation as a separate row
        if len(entity.annotations) > 0:
            for annotation in entity.annotations:
                # Build attribute dict from annotation observations
                attributes = {}
                for obs in annotation.observations:
                    attr_label = (
                        obs.attribute_id.label
                        if hasattr(obs.attribute_id, "label")
                        else str(obs.attribute_id)
                    )
                    attributes[attr_label] = {
                        "value": obs.value,
                        "occurred_at": obs.occurred_at,
                        "confidence": obs.datum_source.confidence,
                    }

                # Add global observations to attributes
                for obs in entity.global_observations:
                    attr_label = (
                        obs.attribute_id.label
                        if hasattr(obs.attribute_id, "label")
                        else str(obs.attribute_id)
                    )
                    attributes[attr_label] = {
                        "value": obs.value,
                        "occurred_at": obs.occurred_at,
                        "confidence": obs.datum_source.confidence,
                    }

                # Get location bounds if location exists
                location_dict = None
                if annotation.location is not None:
                    xmin, ymin, xmax, ymax = annotation.location.bounds
                    location_dict = {
                        "wkt": annotation.location.wkt,
                        "xmin": int(xmin),
                        "ymin": int(ymin),
                        "xmax": int(xmax),
                        "ymax": int(ymax),
                    }

                # Build row data
                row_data = {
                    "entity": {"id": entity.id},
                    "stream": {"id": stream_id if stream_id else "unknown"},
                    "data_sample": {
                        "recorded_at": data_sample.recorded_at,
                        "content_type": data_sample.content_type,
                        "stream_frame_index": data_sample.stream_frame_index,
                        "media_frame_index": data_sample.media_frame_index,
                    },
                    "annotation": {
                        "id": annotation.id,
                    },
                    "attribute": attributes,
                }

                # Add location to annotation if it exists
                if location_dict is not None:
                    row_data["annotation"]["location"] = location_dict

                rows.append(row_data)

        # If entity has no annotations but has global observations, create a row
        elif len(entity.global_observations) > 0:
            # Build attribute dict from global observations only
            attributes = {}
            for obs in entity.global_observations:
                attr_label = (
                    obs.attribute_id.label if hasattr(obs.attribute_id, "label") else str(obs.attribute_id)
                )
                attributes[attr_label] = {
                    "value": obs.value,
                    "occurred_at": obs.occurred_at,
                    "confidence": obs.datum_source.confidence,
                }

            # Build row data without annotation location
            row_data = {
                "entity": {"id": entity.id},
                "stream": {"id": stream_id if stream_id else "unknown"},
                "data_sample": {
                    "recorded_at": data_sample.recorded_at,
                    "content_type": data_sample.content_type,
                    "stream_frame_index": data_sample.stream_frame_index,
                    "media_frame_index": data_sample.media_frame_index,
                },
                "annotation": {
                    # ToDo: How do we get an AnnotationId from global observation
                    # the current setup is "Annotation has-many Observations"
                    # so global_observations, simply dont have an Annotation instance to refer to
                    "id": UUID(int=0),  # Placeholder annotation ID
                },
                "attribute": attributes,
            }

            rows.append(row_data)

        # Note: We no longer create placeholder rows here.
        # Empty entity cases are handled at the ObservationsTable level
        # to allow data_sample expression evaluation without fake entities.
        return rows



import hashlib
import inspect
import json
import threading
import time
import traceback
from functools import lru_cache
from pathlib import Path
from queue import Empty, Queue
from typing import Any, Dict, Iterable, Optional, Tuple, Union
from uuid import UUID

import aiko_services as aiko
from aiko_services import (
    PROTOCOL_PIPELINE,
    ActorTopic,
)
from aiko_services import DataSource as AikoDataSource
from aiko_services import (
    PipelineImpl,
    StreamEvent,
    StreamState,
    compose_instance,
    pipeline_args,
    pipeline_element_args,
)
from pydantic import BaseModel, ConfigDict, Field

from highlighter.agent.capabilities.recorder import (
    Recorder,
    RecordingState,
    RecordMode,
)
from highlighter.client.json_tools import HLJSONEncoder
from highlighter.core.database.database import Database

# ToDo: Move CropArgs out of .dataset
from highlighter.datasets.cropping import CropArgs

__all__ = [
    "ActorTopic",
    "Capability",
    "DataSourceCapability",
    "ContextPipelineElement",
    "EntityUUID",
    "PROTOCOL_PIPELINE",
    "PipelineElement",
    "PipelineImpl",
    "StreamEvent",
    "StreamState",
    "compose_instance",
    "compose_instance",
    "pipeline_args",
    "pipeline_element_args",
]

VIDEO = "VIDEO"
TEXT = "TEXT"
IMAGE = "IMAGE"

EntityUUID = UUID

"""Decouple the rest of the code from aiko.PipelineElement"""
ContextPipelineElement = aiko.ContextPipelineElement
PipelineElement = aiko.PipelineElement

# SEPARATOR = b"\x1c"  # ASCII 28 (File Separator)
SEPARATOR = 28  # ASCII 28 (File Separator)


class _BaseCapability:
    class InitParameters(BaseModel):
        """Populate with init parameter fields"""

        model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)

    class StreamParameters(InitParameters):
        """Populate with stream parameter fields"""

        pass

    @classmethod
    def definition(cls, outputs: Iterable[Tuple[str, str]], **parameters):
        """Generate a dictionary containing information about the Capability

        Args:

        """
        # Get class name
        name = cls.__name__

        # Process InitParameters
        if hasattr(cls, "InitParameters"):
            _ = cls.InitParameters(**parameters)

        # Get input and output parameters from process_frame method signature
        ins = []
        if hasattr(cls, "process_frame"):
            sig = inspect.signature(cls.process_frame)

            # Get input parameters
            for param_name, param in sig.parameters.items():
                if param_name not in ["self", "stream"]:
                    param_type = (
                        param.annotation.__name__ if param.annotation != inspect.Parameter.empty else "Any"
                    )
                    ins.append({"name": param_name, "type": param_type})
        else:
            raise ValueError(f"{cls.__name__} has no `process_frame` function.")

        outs = []
        for n, t in outputs:
            outs.append({"name": n, "type": t})

        # Get module information for deploy
        module_name = cls.__module__

        return {
            "name": name,
            "parameters": parameters,
            "input": ins,
            "output": outs,
            "deploy": {
                "local": {
                    "class_name": name,
                    "module": module_name,
                }
            },
        }

    @property
    @lru_cache(maxsize=1)
    def init_parameters(self):
        # Take init parameters from the agent definition at the
        # agent level and the capability level, with capability parameters
        # overriding agent-level parameters
        parameters = {
            **self.pipeline.definition.parameters,
            **self.definition.parameters,
        }
        return self.InitParameters(**parameters)

    def stream_parameters(self, stream_id):
        stream_parameters = self._get_stream_parameters()
        qualified_stream_parameters = {
            param_name.removeprefix(f"{self.definition.name}."): param_value
            for param_name, param_value in stream_parameters.items()
            if param_name.startswith(self.definition.name)
        }
        # Take stream parameters from the definition and from the stream,
        # with qualified stream parameters (prefixed with the capability name)
        # taking highest precendence.
        # Unqualified stream parameters will not override parameters
        # set on the capability definition, for compatibility with Aiko's parameter precedence.
        parameters = {
            **self.pipeline.definition.parameters,
            **stream_parameters,
            **self.definition.parameters,
            **qualified_stream_parameters,
        }

        # Cache construction of the StreamParameters pydantic model
        # which may be expensive.
        # TODO: Check if serializing to JSON is just as expensive
        # as building pydantic StreamParameters
        if stream_id not in self._cached_stream_parameters:
            self._cached_stream_parameters[stream_id] = (None, None)
        parameters_string = json.dumps(parameters, sort_keys=True, separators=(",", ":"), cls=HLJSONEncoder)
        parameters_hash = hashlib.sha256(parameters_string.encode()).hexdigest()
        cached_hash, cached_params = self._cached_stream_parameters[stream_id]
        if parameters_hash == cached_hash:
            return cached_params
        else:
            validated_params = self.StreamParameters(**parameters)
            self._cached_stream_parameters[stream_id] = (parameters_hash, validated_params)
            return validated_params

    def stop_stream(self, stream, stream_id):
        if stream_id in self._cached_stream_parameters:
            del self._cached_stream_parameters[stream_id]


class _AppendableIterator:
    def __init__(self, max_size: int = 10):
        """Initialize the iterator with an optional `max_size`."""
        self._maxsize = max_size
        self._items = Queue(maxsize=self._maxsize)

    def __iter__(self):
        """Make the class iterable."""
        return self

    def __next__(self):
        """Return the next item in the sequence or raise StopIteration."""
        try:
            item = self._items.get(timeout=10)
        except Empty:
            raise StopIteration
        return item

    def append(self, item):
        """Append an item to the underlying list."""
        self._items.put(item)

    def qsize(self) -> int:
        return self._items.qsize()

    def maxsize(self) -> int:
        return self._maxsize


class Capability(PipelineElement, _BaseCapability):

    class InitParameters(_BaseCapability.InitParameters):
        account_uuid: Optional[UUID] = None
        data_source_uuid: Optional[UUID] = None
        output_filename_template: Optional[str] = None
        output_folder: Optional[str] = None
        recording_state: RecordingState = RecordingState.ON
        record: RecordMode = RecordMode.OFF
        samples_per_file: Optional[int] = None
        seconds_per_file: Optional[float] = None
        database: Optional[Database] = None
        writer_opts: Dict[str, Any] = Field(default_factory=lambda: {})

    class StreamParameters(InitParameters):
        pass

    def recording_enabled(self, stream_id) -> bool:
        parameters = self.stream_parameters(stream_id)
        if not hasattr(parameters, "record") or parameters.record == RecordMode.OFF:
            return False

        required_args = {
            "account_uuid": parameters.account_uuid,
            "data_source_uuid": parameters.data_source_uuid,
        }

        if not all(required_args.values()):
            self.logger.warning(
                f"Recorder args are missing required parameters: {[name for name, value in required_args.items() if not value]}. recording disabled",
                extra={"stream_id": stream_id, "capability_name": self.name},
            )
            return False

        return True

    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)
        self._cached_stream_parameters = {}

    def process_frame(self, stream, *args) -> Tuple[StreamEvent, dict]:
        raise NotImplementedError()

    def start_stream(self, stream, stream_id, use_create_frame=True):
        return StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        _BaseCapability.stop_stream(self, stream, stream_id)
        # Recording cleanup is now handled at the agent level
        return StreamEvent.OKAY, None


# ToDO: Remove
class DataSourceType(BaseModel):
    # class MediaType(str, Enum):
    #    IMAGE = "IMAGE"
    #    TEXT = "TEXT"
    #    VIDEO = "VIDEO"

    media_type: str
    url: str
    id: UUID
    content: Optional[Any] = None

    @classmethod
    def image_iter(cls, images: Iterable[Union[str, Path, bytes]]):
        pass

    @classmethod
    def video_iter(cls, videos: Iterable[Union[str, Path, bytes]]):
        pass

    @classmethod
    def text_iter(cls, tests: Iterable[Union[str, Path, bytes]]):
        pass


class DataSourceCapability(AikoDataSource, _BaseCapability):

    stream_media_type = None

    class InitParameters(Capability.InitParameters):
        rate: Optional[float] = None
        batch_size: int = 1
        data_sources: Optional[str] = None
        file_ids: Optional[Iterable] = None
        task_id: Optional[UUID] = None

    class StreamParameters(InitParameters):
        pass

    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)
        self._dsps: Dict[str, Recorder] = {}
        self._cached_stream_parameters = {}

    def frame_generator(self, stream, pipeline_iter_idx):
        """Produce a batch of frames.

        Args:
            stream: The Stream context
            pipeline_iter_idx: An integer counting the number of times the
                               pipeline has been executed, (ie: process_frame
                               has been called)

        """
        if (
            "disable_create_frame_event" in stream.variables
            and stream.variables["disable_create_frame_event"].is_set()
        ):
            return StreamEvent.STOP, {"diagnostic": "Frame generation disabled"}

        parameters = self.stream_parameters(stream.stream_id)
        batch_size = parameters.batch_size
        task_id = parameters.task_id

        recording_state = parameters.recording_state

        frame_gen_start = time.perf_counter()
        frame_data_batch = {"data_samples": [], "entities": {}}
        for _ in range(batch_size):
            try:
                # self.set_dsp_recording_state(stream)
                sample_start = time.perf_counter()
                data_sample, entities = self.get_next_data_sample(stream)
                sample_elapsed = time.perf_counter() - sample_start
                if sample_elapsed > 0.1:  # Only log if > 100ms
                    self.logger.debug(
                        f"Stream {stream.stream_id} [{self.name}]: get_next_data_sample took {sample_elapsed:.3f}s"
                    )

                frame_data_batch["data_samples"].append(data_sample)
                frame_data_batch["entities"].update(entities)
                self.logger.debug(f"data_sample: {data_sample}, entities: {entities}")
            except StopIteration:
                pass
            except Exception as e:
                frame_data = {"diagnostic": traceback.format_exc()}
                return StreamEvent.ERROR, frame_data

        if not frame_data_batch["data_samples"]:
            return StreamEvent.STOP, {"diagnostic": "All frames generated"}

        # For each pipeline iteration the is a batch of file_ids and frame_ids
        stream.variables["task_id"] = task_id

        frame_gen_elapsed = time.perf_counter() - frame_gen_start
        if frame_gen_elapsed > 0.1:  # Only log if > 100ms
            self.logger.debug(
                f"Stream {stream.stream_id} [{self.name}]: frame_generator took {frame_gen_elapsed:.3f}s for {batch_size} samples"
            )

        return StreamEvent.OKAY, frame_data_batch

    def start_stream(self, stream, stream_id):
        stream.variables["video_capture"] = None
        stream.variables["video_frame_generator"] = None

        stream.variables["disable_create_frame_event"] = threading.Event()

        return super().start_stream(
            stream, stream_id, frame_generator=self.frame_generator, use_create_frame=False
        )

    def stop_stream(self, stream, stream_id):
        if "disable_create_frame_event" in stream.variables:
            stream.variables["disable_create_frame_event"].set()
        return super().stop_stream(stream, stream_id)

    def get_next_data_sample(self, stream):
        raise NotImplementedError()

    def process_frame(
        self, stream, data_samples, entities: Optional[Dict] = None
    ) -> Tuple[StreamEvent, Dict]:
        return StreamEvent.OKAY, {
            "data_samples": data_samples,
            "entities": entities if entities is not None else {},
        }

    def using_hl_data_scheme(self, stream) -> bool:
        return "hl_source_data" in stream.variables



"""Utilities that depend on onnxruntime (ORT)."""

import functools
import logging
from pathlib import Path

from highlighter.core.cuda import is_nvidia_smi_available

LOGGER = logging.getLogger(__name__)


@functools.lru_cache(maxsize=1)
def is_cuda_provider_available() -> bool:
    """Return True if onnxruntime exposes CUDAExecutionProvider."""
    try:
        import onnxruntime as ort
    except Exception:
        return False

    try:
        providers = ort.get_available_providers()
    except Exception:
        return False
    return "CUDAExecutionProvider" in providers


def is_cuda_available() -> bool:
    """Return True when ORT has CUDA provider; logs if only NVIDIA tooling is found."""
    if is_cuda_provider_available():
        return True
    if is_nvidia_smi_available():
        LOGGER.debug("NVIDIA tooling detected but CUDAExecutionProvider is unavailable in onnxruntime.")
    return False


def download_model(
    url: str,
    path: str | Path,
    *,
    timeout: int = 30,
    check_cached: bool = False,
    show_progress: bool = False,
) -> Path:
    """Download a model artifact from URL with status checks and temp-file swap."""
    dest = Path(path)
    if check_cached and dest.exists():
        return dest

    dest.parent.mkdir(parents=True, exist_ok=True)
    tmp_path = dest.with_suffix(dest.suffix + ".partial")
    try:
        import requests
    except ImportError as exc:
        raise ImportError("requests is required to download ONNX models") from exc

    if show_progress:
        LOGGER.info("Downloading model from %s to %s", url, dest)
    else:
        LOGGER.debug("Downloading model from %s to %s", url, dest)

    try:
        with requests.get(url, stream=True, timeout=timeout) as response:
            response.raise_for_status()
            total_length = int(response.headers.get("content-length", 0))
            chunk_iter = response.iter_content(chunk_size=1024 * 1024)

            if show_progress:
                try:
                    import tqdm  # type: ignore

                    total = None
                    if total_length:
                        total = max(1, total_length // (1024 * 1024))
                    chunk_iter = tqdm.tqdm(
                        chunk_iter,
                        total=total,
                        bar_format="{l_bar}{bar:10}",
                    )
                except Exception as exc:
                    LOGGER.debug("tqdm unavailable; proceeding without progress bar: %s", exc)

            with tmp_path.open("wb") as f:
                for chunk in chunk_iter:
                    if chunk:
                        f.write(chunk)
            tmp_path.replace(dest)
    except Exception:
        if tmp_path.exists():
            tmp_path.unlink()
        raise

    return dest



from datetime import datetime
from pathlib import Path

import click
import yaml
from pydantic import BaseModel

from highlighter.cli.common import CommonOptions

from ..client import get_latest_assessments_gen, multithread_graphql_file_download
from ..client.ingestion_setup import bootstrap_ingestion
from ..client.tasks import add_files_to_order
from ..datasets import (
    SUPPORTED_SPLIT_FNS,
    CocoWriter,
    Dataset,
    DatasetFormat,
    HighlighterAssessmentsWriter,
    YoloWriter,
    get_reader,
    get_split_fn,
)

TASK_CREATION_BATCH_SIZE = 1000


def _download_data_files(client, data_file_ids, data_file_dir):
    if data_file_dir is not None:
        multithread_graphql_file_download(
            client,
            data_file_ids,
            data_file_dir,
        )


def _combine_datasets(ctx):
    client = ctx.obj["client"]
    page_size = ctx.obj["page_size"]
    dataset_splits = ctx.obj["dataset_splits"]

    downloaded = {}
    datasets = []
    for dataset_id, split in dataset_splits:
        if dataset_id in downloaded:
            ds = Dataset(
                annotations_df=downloaded[dataset_id].annotations_df.copy(),
                data_files_df=downloaded[dataset_id].data_files_df.copy(),
            )
        else:
            ds = Dataset.read_from(
                dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
                client=client,
                dataset_id=dataset_id,
                page_size=page_size,
            )
        ds.data_files_df.split = split
        datasets.append(ds)
        downloaded[dataset_id] = ds

    return Dataset.combine(datasets)


@click.group("dataset")
@click.pass_context
def dataset_group(ctx):
    pass


@dataset_group.command(
    "create",
    help="""
Create a Dataset in Highlighter

\b
Note: If using the --dataset-id param. There are no safety rails! That is to
say; the cli will allow duplicate assessments and data_files if you ask it to.

  ie: If dataset 123 has data_file 456 and workflow 789 also has data_file 456 the
  assessments for both will be added to the dataset.

Typically you would use the --dataset-id param if you wanted to combine 2 or
more datasets, or combine some new data from a workflow with an existing dataset.
In which case a new dataset would be created that simply has all assessments
appended together.
        """,
)
@click.option(
    "-n",
    "--name",
    type=str,
    required=True,
)
@click.option(
    "-w",
    "--workflow-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-oid",
    "--object-class-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-uid",
    "--user-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-did",
    "--dataset-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
    help=(
        "Add assessments from existing dataset(s), see Note in help string "
        " for more context and a warrning."
    ),
)
@click.option(
    "--split-type",
    type=click.Choice(SUPPORTED_SPLIT_FNS.keys()),
    required=False,
    default="RandomSplitter",
    show_default=True,
)
@click.option(
    "--split-seed",
    type=int,
    required=False,
    default=42,
    show_default=True,
)
@click.option(
    "--split-frac",
    type=click.Tuple([str, float]),
    required=False,
    multiple=True,
    default=[("data", 1.0)],
    show_default=True,
)
@click.option(
    "-c",
    "--classes-of-interest",
    type=str,
    required=False,
    default=None,
    multiple=True,
    help="Object class uuids to consider when validating splits",
)
@click.option(
    "--page-size",
    type=int,
    required=False,
    default=200,
    help="Page Size used when fetching Assessments from Highlighter",
)
@click.pass_context
def create(
    ctx,
    name,
    workflow_id,
    object_class_id,
    user_id,
    dataset_id,
    split_type,
    split_seed,
    split_frac,
    classes_of_interest,
    page_size,
):
    client = ctx.obj["client"]

    status = 0  # completed
    query_args = dict(
        workflowId=workflow_id,
        objectClassId=object_class_id,
        userId=user_id,
        status=status,
        annotationStatus="HAS_ANNOTATIONS",
    )

    def not_empty(v):
        if v is None:
            return False
        if v == ():
            return False
        if v == []:
            return False
        return True

    query_args = {k: v for k, v in query_args.items() if not_empty(v)}

    datasets = []
    if query_args:
        latest_subs_gen = get_latest_assessments_gen(
            client,
            page_size=page_size,
            **query_args,
        )

        reader = get_reader("highlighter_assessments")(latest_subs_gen)
        ds = Dataset.load_from_reader(reader)
        datasets.append(ds)

    dataset_description_fields = []
    if dataset_id:
        # add to query_args here because datasetId is not a valid
        # arg to get_latest_assessments_gen, but we still want it
        # to appear in the Highlighter Dataset description.
        query_args["datasetId"] = dataset_id
        dataset_list = [
            Dataset.read_from(
                dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
                client=client,
                dataset_id=ds_id,
            )
            for ds_id in dataset_id
        ]

        datasets.extend(dataset_list)

        def get_dataset_url(id, client=client):
            return client.endpoint_url.replace("graphql", f"datasets/{id}")

        def get_dataset_name(id, client=client):
            class DatasetNameOnly(BaseModel):
                name: str

            name = client.dataset(
                return_type=DatasetNameOnly,
                id=id,
            ).name
            return name.replace("_", "\\_")

        base_dataset_links = [
            f"  - [{id}]({get_dataset_url(id)}): {get_dataset_name(id)}" for id in dataset_id
        ]
        base_dataset_links_str = "\n".join(base_dataset_links)
        dataset_description_fields.append(("Base Datasets", base_dataset_links_str))

    dataset = Dataset.combine(datasets)
    if classes_of_interest:
        adf = dataset.annotations_df
        ddf = dataset.data_files_df

        adf = adf[adf.value.isin(classes_of_interest)]
        ddf = ddf[ddf.data_file_id.isin(adf.data_file_id.unique())]
        dataset.annotations_df = adf
        dataset.data_files_df = ddf

    split_names = [s[0] for s in split_frac]
    fracs = [s[1] for s in split_frac]
    splitter = get_split_fn(split_type)(split_seed, fracs, split_names)

    dataset.apply_split(splitter)

    split_args = {
        "type": splitter.__class__.__name__,
        "seed": splitter.seed,
        "fracs": splitter.fracs,
        "names": splitter.names,
    }

    split_str = yaml.safe_dump(split_args)
    split_str = f"<pre>" + split_str + "</pre> \n"

    query_str = yaml.safe_dump(query_args)
    query_str = f"<pre>" + query_str + "</pre> \n"

    # List of tuples with e[0] being the heading and e[1]
    # being the content. This will be rendered as Markdown
    # in the publish_to_highlighter function.
    dataset_description_fields.extend(
        [
            ("Query", query_str),
            ("Split", split_str),
        ]
    )

    dataset.publish_to_highlighter(
        client,
        name,
        dataset_description_fields=dataset_description_fields,
        split_fracs={s: f for s, f in zip(split_names, fracs)},
    )


def _resolve_dataset_name(ctx, local_name, default_name):
    """
    Helper to enforce a single invocation of the --name/-n flag across
    `hl dataset upload FORMAT` commands.
    """
    if ctx.obj.get("dataset_name") and local_name:
        raise click.UsageError("The --name/-n flag should be used no more than once.")

    return ctx.obj.get("dataset_name") or local_name or default_name


def _upload_impl(ctx, ds, name, class_map=None):
    """
    Format-agnostic implementation of dataset upload logic.

    Args:
        ctx: context from click
        ds: a loaded Dataset object
        name: name of the dataset to create on the server
        class_map: optional dict of class_id -> class_name derived from readers.
    """
    obj_classes = set()
    if class_map:
        obj_classes.update(class_map.values())
    elif "value" in ds.annotations_df.columns:
        obj_classes.update(v for v in ds.annotations_df["value"].dropna().unique() if isinstance(v, str))

    # TODO: roll back workflow and data source creation on failure
    # create the project, workflow, etc. on the server
    client = ctx.obj["client"]
    infrastructure = bootstrap_ingestion(client, name, list(obj_classes))

    click.echo("Uploading images to new data source...")
    ds.upload_data_files(client, data_source_id=infrastructure.data_source_id, progress=True)

    # create task/assessment placeholders and get the mapping of
    # data_file_id -> assessment_id
    file_ids = ds.data_files_df["data_file_id"].dropna().unique().tolist()
    # ensure IDs are passed as strings to the API, handling floats if necessary
    file_ids = [str(int(f)) if isinstance(f, (int, float)) else str(f) for f in file_ids]

    click.echo(f"Preparing assessment submissions for {len(file_ids)} files...")
    task_map = {}
    for i in range(0, len(file_ids), TASK_CREATION_BATCH_SIZE):
        batch = file_ids[i : i + TASK_CREATION_BATCH_SIZE]
        task_map.update(add_files_to_order(client, infrastructure.order_id, batch))

    # wrap each image from our dataset and its associated attributes in a
    # Highlighter Annotation, and push it to web as a task
    HighlighterAssessmentsWriter(client, workflow_id=infrastructure.workflow_id, task_lookup=task_map).write(
        ds
    )

    # homogenise any distinct split names
    ds.data_files_df["split"] = "data"
    # create new dataset on web comprising all of the
    ds.publish_to_highlighter(client, f"{name}_{datetime.now().strftime('%H%M%S')}", split_fracs={"data": 1})
    click.echo(f"\nUpload complete.")


@dataset_group.group("upload", subcommand_metavar="FORMAT")
@click.option(
    "--name", "-n", required=False, default=None, help="The name of the new dataset to create in Highlighter."
)
@click.pass_context
def upload(ctx, name):
    """
    Upload a local object detection dataset to Highlighter.

    This command requires a FORMAT argument (one of 'yolo' or 'coco')
    in order to parse your dataset correctly.
    """
    ctx.ensure_object(dict)
    ctx.obj["dataset_name"] = name


# TODO:
# Two new options:
# - Provide a list of split/partition names defined in the YAML to ingest as a single dataset.
# - If previous option disabled, users can manually qualify 'image dir, label dir' pairs.
#   - We suggest this because at present, the standard YOLO schema for YAML files denotes each
#     split by its *images* directory--no mention of that split's corresponding label directory
#     exists.


@upload.command("yolo")
@click.argument("yaml_path", type=click.Path(exists=True, path_type=Path, resolve_path=True, dir_okay=False))
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="The name of the new dataset to create in Highlighter.",
)
@click.option(
    "--working-directory",
    "-d",
    "working_dir",
    type=click.Path(exists=True, path_type=Path, resolve_path=True, file_okay=False),
    default=Path.cwd(),
    help="Base directory from which to interpret relative paths in the YAML file.",
)
@click.pass_context
def upload_yolo(ctx, yaml_path, working_dir, name):
    """
    Upload a YOLO-formatted dataset.

    YAML_PATH: Path to the data.yaml file definition.
    """
    dataset_name = _resolve_dataset_name(ctx, name, "YOLO Dataset")

    reader = get_reader("yolo")(yaml_path, working_dir)
    ds = Dataset.load_from_reader(reader)

    _upload_impl(ctx, ds, name=dataset_name, class_map=reader.class_map)


@upload.command("coco")
@click.argument(
    "json_path",
    required=False,
    type=click.Path(exists=True, path_type=Path, resolve_path=True, dir_okay=False),
)
@click.argument(
    "image_dir",
    required=False,
    type=click.Path(exists=True, path_type=Path, resolve_path=True, file_okay=False),
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="The name of the new dataset to create in Highlighter.",
)
@click.option(
    "--input",
    "-i",
    "additional_splits",
    type=(
        click.Path(exists=True, path_type=Path, resolve_path=True, dir_okay=False),
        click.Path(exists=True, path_type=Path, resolve_path=True, file_okay=False),
    ),
    multiple=True,
    required=False,
    help="""
    Submit *additional* (json_path, image_dir) pairs for ingestion (i.e., prepend --input/-i to every pair after the first).
    """,
)
@click.option(
    "--bbox-only",
    is_flag=True,
    help="Ignore segmentation data and only import bounding boxes.",
)
@click.pass_context
def upload_coco(ctx, json_path, image_dir, name, additional_splits, bbox_only):
    """
    Upload a COCO-formatted dataset.

    You can upload multiple JSON files (e.g. train and val) by repeating the --input flag.

    \b
    Example:

        hl dataset upload -n my_ds coco train.json ./train_imgs -i val.json ./val_imgs
    """
    dataset_name = _resolve_dataset_name(ctx, name, "COCO Dataset")

    # collect all pairs of (json_path, image_dir)
    all_pairs = list(additional_splits)
    if json_path and image_dir:
        all_pairs.append((json_path, image_dir))
    elif json_path or image_dir:
        raise click.UsageError("Must provide both JSON_PATH and IMAGE_DIR if using positional arguments.")

    if not all_pairs:
        raise click.UsageError(
            "No datasets provided. Please specify JSON_PATH and IMAGE_DIR or use --input/-i."
        )

    datasets = []
    class_maps = []
    for j_path, i_dir in all_pairs:
        click.echo(f"Loading {j_path}...")
        reader = get_reader("coco")(Path(j_path), image_dir=Path(i_dir), bbox_only=bbox_only)
        class_maps.append(reader.class_map)
        datasets.append(Dataset.load_from_reader(reader))

    if not datasets:
        raise click.UsageError("No datasets loaded.")

    # combine all loaded COCO parts into one Dataset,
    # and all class maps into one map
    combined_ds = Dataset.combine(datasets)
    combined_class_map = {}
    for m in class_maps:
        combined_class_map.update(m)

    _upload_impl(ctx, combined_ds, dataset_name, class_map=combined_class_map)


@click.group("read")
@click.option(
    "-i",
    "--dataset-ids",
    type=str,
    required=False,
    multiple=True,
    default=[],
    show_default=True,
    help=" int or int:str --> ID or ID:SPLIT",
)
@click.option(
    "--page-size",
    type=int,
    required=False,
    default=None,
    show_default=True,
)
@click.pass_context
def read_group(ctx, dataset_ids, page_size):
    """Read Datasets from Highlighter and write to local disk.

    If page_size is None, the page_size is set from:
        HighlighterRuntimeConfig.default_pagination_page_size

    For help with each specific output format use:

    \b
      hl dataset read coco|yolo|... --help

    \b
    By default Datasets spits are all set to 'data' to assign
    a split to a dataset use --dataset-id ID:SPLIT

    \b
    Example:
      read datasets 111 and 222, assign 111 to 'train' split and
      222 to 'val' split then write to yolo detection format

    \b
      hl dataset read -i 111:train -i 204:val yolo -t detection my_output
    """
    ctx.obj["page_size"] = page_size if page_size is not None else ctx.obj["hl_cfg"].pagination_page_size

    if all([":" in i for i in dataset_ids]):
        dataset_splits = [i.split(":") for i in dataset_ids]
        dataset_splits = [(int(id), split) for id, split in dataset_splits]
    else:
        dataset_splits = [(int(i), "data") for i in dataset_ids]
    ctx.obj["dataset_splits"] = dataset_splits


@read_group.command("hdf")
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def write_hdf(ctx, annotations_dir, data_file_dir):
    filestem = "-".join([str(i[0]) for i in ctx.obj["dataset_splits"]])
    filename = f"{filestem}.hdf"

    hdf_path = annotations_dir / filename
    combined_dataset = _combine_datasets(ctx)
    combined_dataset.write_hdf(hdf_path)

    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


@read_group.command("json")
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def write_json(ctx, annotations_dir, data_file_dir):
    filestem = "-".join([str(i[0]) for i in ctx.obj["dataset_splits"]])
    filename = f"{filestem}.json"

    json_path = annotations_dir / filename
    combined_dataset = _combine_datasets(ctx)
    combined_dataset.write_json(json_path)

    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


@read_group.command(CocoWriter.format_name)
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def write_coco(ctx, annotations_dir, data_file_dir):
    writer = CocoWriter(annotations_dir)

    combined_dataset = _combine_datasets(ctx)

    writer.write(combined_dataset)
    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


@read_group.command(YoloWriter.format_name)
@click.argument("output_dir", type=str)
@CommonOptions.data_file_dir
@click.option(
    "--task",
    "-t",
    type=click.Choice([member.value for member in YoloWriter.TASK]),
    default=YoloWriter.TASK.DETECT,
)
@click.pass_context
def write_yolov8(ctx, output_dir, data_file_dir, task):
    image_cache_dir = Path(".") if data_file_dir is None else data_file_dir
    writer = YoloWriter(output_dir, image_cache_dir, task=task)

    combined_dataset = _combine_datasets(ctx)
    writer.write(combined_dataset)

    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


dataset_group.add_command(read_group)



import json
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Dict, List, Optional, Union
from uuid import UUID
from warnings import warn

import yaml
from pydantic import BaseModel, BeforeValidator, ConfigDict, Field, field_validator
from typing_extensions import Annotated

from highlighter.datasets.cropping import CropArgs

from ..core.gql_base_model import GQLBaseModel
from .gql_client import HLClient
from .training_runs import TrainingRunArtefactTypeEnum

__all__ = [
    "get_training_config",
    "EntityAttributeValueTypeEnum",
    "EntityAttributeValue",
    "EntityAttribute",
    "ModelOutputType",
    "ModelInputs",
    "ModelSchemaType",
    "DatasetType",
    "ArtefactInfo",
    "TaxonPart",
    "TrainingConfigType",
]


def validate_uuid(v, key):
    validated = None
    if isinstance(v, UUID):
        validated = str(v)
    else:
        try:
            UUID(v)
            validated = v
        except:
            msg = f"'{key}' must be valid UUID, got: {v}"
            raise ValueError(msg)
    return validated


class EntityAttributeValueTypeEnum(str, Enum):
    boolean = "boolean"
    enum = "enum"
    integer = "integer"
    decimal = "decimal"
    string = "string"
    geometry = "geometry"
    array = "array"


class TaxonPart(BaseModel, extra="forbid"):
    entity_attribute_enum_id: Optional[Union[UUID, str]] = None
    entity_attribute_enum_value: Optional[str] = None
    entity_attribute_id: Union[UUID, str]
    entity_attribute_name: str
    entity_attribute_value_type: Optional[EntityAttributeValueTypeEnum] = None

    @field_validator("entity_attribute_enum_id")
    @classmethod
    def validate_entity_attribute_enum_id(cls, v):
        if v is None:
            return True
        else:
            return str(validate_uuid(v, "entity_attribute_enum_id"))

    @field_validator("entity_attribute_id")
    @classmethod
    def validate_entity_attribute_id(cls, v):
        return str(validate_uuid(v, "entity_attribute_id"))

    @property
    def value_type(self):
        return self.entity_attribute_value_type


class ModelOutputType(BaseModel, extra="forbid"):
    head: int
    position: int
    conf_thresh: Annotated[float, Field(ge=0.0, lt=1.0)]
    taxon: List[TaxonPart]

    def _validate_one_taxon(self, property_name):
        if len(self.taxon) > 1:
            raise ValueError(
                f"Can only use ModelOutputType.{property_name} if there is a single taxa in self.taxon"
            )

    @property
    def entity_attribute_enum_value(self):
        self._validate_one_taxon("entity_attribute_enum_value")
        return self.taxon[0].entity_attribute_enum_value

    @property
    def entity_attribute_enum_id(self) -> Optional[Union[UUID, str]]:
        self._validate_one_taxon("entity_attribute_enum_id")
        return self.taxon[0].entity_attribute_enum_id

    @property
    def entity_attribute_id(self) -> Union[UUID, str]:
        self._validate_one_taxon("entity_attribute_id")
        return self.taxon[0].entity_attribute_id

    @property
    def entity_attribute_name(self) -> str:
        self._validate_one_taxon("entity_attribute_name")
        return self.taxon[0].entity_attribute_name

    @property
    def entity_attribute_value_type(self) -> Optional[EntityAttributeValueTypeEnum]:
        self._validate_one_taxon("entity_attribute_value_type")
        return self.taxon[0].entity_attribute_value_type


class EntityAttribute(BaseModel, extra="forbid"):
    entity_attribute_id: Union[UUID, str]
    entity_attribute_name: str

    @field_validator("entity_attribute_id")
    @classmethod
    def validate_entity_attribute_id(cls, v):
        return str(validate_uuid(v, "entity_attribute_id"))


class EntityAttributeValue(BaseModel, extra="forbid"):
    entity_attribute_enum_id: Optional[Union[UUID, str]] = None
    entity_attribute_enum_value: Optional[str] = None
    entity_attribute_id: Union[UUID, str]
    entity_attribute_name: str
    entity_attribute_value_type: EntityAttributeValueTypeEnum
    entity_attribute_value: Any = None

    @field_validator("entity_attribute_enum_id")
    @classmethod
    def validate_entity_attribute_enum_id(cls, v):
        if v is None:
            return v
        return str(validate_uuid(v, "entity_attribute_enum_id"))

    @field_validator("entity_attribute_id")
    @classmethod
    def validate_entity_attribute_id(cls, v):
        return str(validate_uuid(v, "entity_attribute_id"))

    @property
    def value_type(self):
        return self.entity_attribute_value_type

    @property
    def value(self):
        if self.value_type == EntityAttributeValueTypeEnum.enum:
            return self.entity_attribute_enum_id
        else:
            return self.entity_attribute_value


class ModelInputs(BaseModel, extra="forbid"):
    entity_attributes: List[EntityAttribute]
    filters: List[List[EntityAttributeValue]]


class ModelSchemaType(BaseModel, extra="forbid"):
    model_config = ConfigDict(protected_namespaces=())

    model_inputs: ModelInputs
    model_outputs: List[ModelOutputType]

    @field_validator("model_outputs")
    @classmethod
    def validate_model_output_positions(cls, model_outputs):
        if len(model_outputs) == 0:
            #####################################################################
            # ❗❗❗ Remove when all inference capabilities are using TrainingConfig
            # as oposed schema's tied up in ManifestV1
            #####################################################################
            return model_outputs

        model_outputs = sorted(model_outputs, key=lambda o: (o.head, o.position))
        if (model_outputs[0].position != 0) and (model_outputs[0].head != 0):
            raise ValueError("model_outputs ids must start at 0")

        num_heads = len(set([o.head for o in model_outputs]))
        for head_idx in range(num_heads):
            for pos_idx, o in enumerate([_o for _o in model_outputs if _o.head == head_idx]):
                if head_idx != o.head:
                    raise ValueError("model_output.head values must start at 0 and be contigious")
                if pos_idx != o.position:
                    raise ValueError("model_output.position values must start at 0 and be contigious")

        return model_outputs

    def get_input_select_attribute_ids(self) -> List[str]:
        return [str(i.entity_attribute_id) for i in self.model_inputs.entity_attributes]

    def get_input_filter_attribute_ids(self) -> List[str]:
        filter_attr_ids = []
        for input_filter in self.model_inputs.filters:
            filter_attr_ids.append([str(i.entity_attribute_id) for i in input_filter])
        return filter_attr_ids

    def get_input_filter_attribute_values(self) -> List[Any]:
        filter_values = []
        for input_filter in self.model_inputs.filters:
            filter_values.append([i.value for i in input_filter])

        filter_values = [v if len(v) > 0 else None for v in filter_values]
        return filter_values

    def get_head_model_outputs(self, head: int):
        model_outputs = [m for m in self.model_outputs if m.head == head]
        return model_outputs

    def get_head_output_attribute_ids(self, head: int) -> List[str]:
        head_model_outputs = self.get_head_model_outputs(head)
        return [str(m.entity_attribute_id) for m in head_model_outputs]

    def get_head_output_attribute_names(self, head: int) -> List[str]:
        head_model_outputs = self.get_head_model_outputs(head)
        return [str(m.entity_attribute_name) for m in head_model_outputs]

    def get_head_output_attribute_enum_ids(self, head: int) -> List[str]:
        head_model_outputs = self.get_head_model_outputs(head)
        return [str(m.entity_attribute_enum_id) for m in head_model_outputs]

    def get_head_output_attribute_enum_values(self, head: int) -> List[str]:
        head_model_outputs = self.get_head_model_outputs(head)
        return [str(m.entity_attribute_enum_value) for m in head_model_outputs]

    def get_head_output_value_type(self, head: int) -> Optional[EntityAttributeValueTypeEnum]:
        """Will return None if the entity_attribute_value_type is also None
        because it is an Optional parameter of ModelOutputType
        """
        head_model_outputs = self.get_head_model_outputs(head)
        head_output_value_types = {m.entity_attribute_value_type for m in head_model_outputs}
        assert len(head_output_value_types) == 1
        return head_output_value_types.pop()


def _validate_hash_signature_not_none(v, args) -> str:
    assert (
        v is not None
    ), f"dataset({args.data['id']}) hash_signature is None. This is likely becuase it has not been locked"
    return v


class DatasetType(BaseModel, extra="forbid"):
    id: int
    hash_signature: Annotated[str, BeforeValidator(_validate_hash_signature_not_none)]


class ArtefactInfo(GQLBaseModel):
    """camelCase dict keys are intended as these args will be passed to a graphql mutation"""

    type: TrainingRunArtefactTypeEnum
    training_run_id: int
    artefact_path: str
    checkpoint: str
    inference_config: Dict
    training_config: Dict

    def to_dict(self) -> dict:
        returnDict = self.model_dump()
        returnDict["type"] = str(self.type.name)
        return returnDict


class TrainingConfigType(BaseModel, extra="forbid"):
    class TrainerType(str, Enum):
        YOLO_DET = "yolo-det"
        YOLO_SEG = "yolo-seg"
        YOLO_CLS = "yolo-cls"

    # Configure the BaseModel to allow us to populate the BaseModel using
    # alias or the name
    model_config = ConfigDict(populate_by_name=True, protected_namespaces=())

    base_conf: Optional[str] = None
    config_overrides: Dict
    data: Dict[str, List[DatasetType]]

    # model_config is a reserved name in pydantic so we've renamed to base_config
    trainer_arguments: Dict = Field(..., alias="model_arguments")
    trainer_config: Dict = Field(..., alias="model_config")
    input_output_schema: ModelSchemaType = Field(..., alias="model_schema")

    training_run_id: int
    research_plan_id: Optional[int] = None
    experiment_id: Optional[int] = None

    trainer_type: Optional[TrainerType] = None
    head_idx: int = 0
    crop_args: CropArgs = CropArgs()
    oversample: bool = False

    @property
    def model_schema(self):
        warn("TrainingConfigType.model_schema->input_output_schema", DeprecationWarning, stacklevel=2)
        return self.input_output_schema

    def get_dataset_ids(self, purpose: str):
        purposes = ["train", "test", "dev"]
        if purpose not in purposes:
            raise KeyError(f"Invalid dataset purpose, expected one of {purposes} " f"got: {purpose}")

        return [d.id for d in self.data.get(f"datasets_{purpose}", [])]

    def get_dataset_purpose_id_lookup(self) -> Dict[str, List[int]]:
        purposes = ["train", "test", "dev"]
        return {p: self.get_dataset_ids(p) for p in purposes}

    @classmethod
    def _from_dict(cls, data):
        RENAMED_FIELDS = (
            ("model_arguments", "trainer_arguments"),
            ("model_config", "trainer_config"),
            ("model_schema", "input_output_schema"),
        )

        for old_name, new_name in RENAMED_FIELDS:
            if old_name in data:
                val = data.pop(old_name)
                data[new_name] = val
        return cls(**data)

    @classmethod
    def from_json(cls, json_path: Path):
        with json_path.open("r") as f:
            data = json.load(f)
        return cls._from_dict(data)

    @classmethod
    def from_yaml(cls, yaml_path: Path):
        with yaml_path.open("r") as f:
            data = yaml.safe_load(f)
        return cls._from_dict(data)

    def dump(self, fmt: str, path: Union[str, Path]):
        path = Path(path)
        dumpers = {
            "yaml": self._dump_yaml,
            "json": self._dump_json,
        }
        dumper = dumpers[fmt]
        dumper(path)

    def _dump_yaml(self, path: Path):

        with path.open("w") as f:
            yaml.safe_dump(self.model_dump(), f, default_flow_style=False)

    def _dump_json(self, path: Path):
        with path.open("w") as f:
            json.dump(self.model_dump(), f, indent=2)

    @property
    def evaluation_id(self) -> Optional[int]:
        return self.research_plan_id

    @classmethod
    def from_highlighter(cls, client: HLClient, training_run_id: int):
        return get_training_config(client, training_run_id)


def _update_model_outputs_to_use_taxon(model_outputs):
    updated_model_outputs = []
    for mo in model_outputs:
        if "taxon" in mo:
            updated_model_outputs.append(mo)
        else:
            taxa = {}
            for key in (
                "entity_attribute_id",
                "entity_attribute_name",
                "entity_attribute_value_type",
                "entity_attribute_enum_id",
                "entity_attribute_enum_value",
            ):
                taxa[key] = mo.pop(key, None)
            mo["taxon"] = [taxa]
            updated_model_outputs.append(mo)
    return updated_model_outputs


def get_training_config(
    client: HLClient,
    training_run_id: int,
) -> TrainingConfigType:
    class TrainingRunType(GQLBaseModel):
        training_config: dict

    result = client.trainingRun(
        return_type=TrainingRunType,
        id=training_run_id,
    )

    # ToDo: Intermediate backward compat code mapping single attr model model_outputs
    # to taxon model_outputs as per https://github.com/silverpond/highlighter-web/pull/44
    # when this is on master we can remove this code.
    training_config = result.training_config
    training_config["model_schema"]["model_outputs"] = _update_model_outputs_to_use_taxon(
        training_config["model_schema"]["model_outputs"],
    )
    ### END ###

    val = result.training_config["config_overrides"].pop("crop_args", {})
    result.training_config["crop_args"] = val
    result.training_config["model_config"].pop("crop_args", {})

    if "oversample" in result.training_config["config_overrides"]:
        result.training_config["oversample"] = result.training_config["config_overrides"].pop("oversample")

    return TrainingConfigType(**result.training_config)


# Custom representer for the Enum
def enum_representer(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data.value)


yaml.SafeDumper.add_representer(EntityAttributeValueTypeEnum, enum_representer)
yaml.SafeDumper.add_representer(TrainingConfigType.TrainerType, enum_representer)



import json
import logging
import os
import threading
import time
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Optional, Union

import boto3
import yaml
from gql import Client as GQLClient
from gql import gql
from gql.transport.requests import RequestsHTTPTransport
from pydantic import BaseModel

import highlighter.core.decorators as decorators
from highlighter.client._colors import ColoredString
from highlighter.client.json_tools import HLJSONEncoder

from ..core import snake_to_camel
from .gql_base import (
    Line,
    get_all_mutations,
    get_all_queries,
    get_gql_obj,
    get_gql_return_type,
    get_gql_schema,
    return_type_formatting,
    to_gql_type,
    to_python_type,
)

LOGGER = logging.getLogger(".".join([__name__, "HLClient"]))

cs = ColoredString()


KEY_API_TOKEN = "api_token"  # nosec hardcoded_password_string
KEY_ENDPOINT_URL = "endpoint_url"
KEY_CLOUD = "cloud"

KEY_AWS_ACCESS_KEY_ID = "aws_access_key_id"
KEY_AWS_SECRET_ACCESS_KEY = "aws_secret_access_key"  # nosec hardcoded_password_string
KEY_AWS_REGION = "aws_default_region"

ENV_HL_WEB_GRAPHQL_API_TOKEN = "HL_WEB_GRAPHQL_API_TOKEN"  # nosec hardcoded_password_string
ENV_HL_WEB_GRAPHQL_ENDPOINT = "HL_WEB_GRAPHQL_ENDPOINT"
ENV_AWS_ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"
ENV_AWS_SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"  # nosec hardcoded_password_string
ENV_AWS_DEFAULT_REGION = "AWS_DEFAULT_REGION"

ENV_HL_DEFAULT_PROFILE = "HL_DEFAULT_PROFILE"
ENV_HL_PROFILES_YAML = "HL_PROFILES_YAML"
ENV_HL_GQL_TIMEOUT_SEC = "HL_GQL_TIMEOUT_SEC"

CONST_HLCLIENT_GQL_TIMEOUT_SEC = os.environ.get(ENV_HL_GQL_TIMEOUT_SEC, 60)
CONST_GRAPHQL_DEFAULT_PROFILE = os.environ.get(ENV_HL_DEFAULT_PROFILE, None)
CONST_DEFAULT_GRAPHQL_PROFILES_YAML = Path.home() / ".highlighter-profiles.yaml"

CONST_GRAPHQL_PROFILES_YAML = os.environ.get(ENV_HL_PROFILES_YAML, CONST_DEFAULT_GRAPHQL_PROFILES_YAML)

EXAMPLE_PROFILE = {
    "my-first-profile": {
        KEY_ENDPOINT_URL: "https://<client-account>.highlighter.ai/graphql",
        KEY_API_TOKEN: "123...abc",
    }
}


class S3Creds(BaseModel):
    __type__: str = "aws-s3"
    type: Literal[__type__] = __type__
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_default_region: str

    def as_environment_variables(self) -> Dict[str, str]:
        return {
            ENV_AWS_ACCESS_KEY_ID: self.aws_access_key_id,
            ENV_AWS_SECRET_ACCESS_KEY: self.aws_secret_access_key,
            ENV_AWS_DEFAULT_REGION: self.aws_default_region,
        }


def validate_cloud_creds(cloud_creds_dict):
    valid_cloud_cred_types = [S3Creds]
    for t in valid_cloud_cred_types:
        try:
            creds = t(**cloud_creds_dict)
            return creds
        except:
            continue  # nosec try_except_continue
    raise ValueError(f"Not a valid cloud credential: {cloud_creds_dict}")


def get_credentials_from_profiles_yaml(
    profile,
    profiles_path=CONST_GRAPHQL_PROFILES_YAML,
):
    if not Path(profiles_path).exists():
        raise FileNotFoundError()

    with open(profiles_path, "r") as f:
        creds = yaml.safe_load(f).get(profile)

    if creds is None:
        raise KeyError(f"Profile '{profile}' could not be found in '{profiles_path}'")

    # Used to access private cloud storage outside of Highlighter S3
    cloud_cred_dicts = creds.get(KEY_CLOUD, None)
    if cloud_cred_dicts is not None:
        cloud_creds = {}
        for cred_dict in cloud_cred_dicts:
            _creds = validate_cloud_creds(cred_dict)
            cloud_creds[_creds.type] = _creds
    else:
        cloud_creds = None

    return creds[KEY_API_TOKEN], creds[KEY_ENDPOINT_URL], cloud_creds


class HLClient(object):
    EXECUTION_RETRYS = 10
    _instance = None

    def __init__(self, client: GQLClient, cloud_creds: Optional[Dict[str, str]] = None):
        self._client = client
        # Lazy-load schema on first use to prevent network request until needed
        self._schema = None
        self._schema_lock = threading.Lock()
        self._is_closed = False

        # Used to access private cloud storage outside of Highlighter S3
        self.cloud_creds = cloud_creds
        HLClient._instance = self

    def _ensure_schema(self):
        """Fetch and cache the GraphQL schema once, thread-safely."""
        if self._schema is None:
            with self._schema_lock:
                if self._schema is None:  # double-checked locking
                    self._schema = get_gql_schema(self)

    def refresh_schema(self):
        """Force a re-introspection of the schema (e.g., after server upgrade)."""
        with self._schema_lock:
            self._schema = get_gql_schema(self)

    def set_schema_for_tests(self, schema):
        """Allow tests to inject a stub schema without hitting the network."""
        with self._schema_lock:
            self._schema = schema

    @property
    def schema(self):
        """Public, read-only accessor that ensures the schema is loaded."""
        self._ensure_schema()
        return self._schema

    @classmethod
    def from_profile(cls, profile: str, profiles_path=None) -> "HLClient":
        profiles_path = profiles_path or CONST_DEFAULT_GRAPHQL_PROFILES_YAML
        api_token, endpoint_url, cloud_creds = get_credentials_from_profiles_yaml(
            profile, profiles_path=profiles_path
        )

        return cls.from_credential(api_token, endpoint_url, cloud_creds=cloud_creds)

    @classmethod
    def from_credential(
        cls,
        api_token: str,
        endpoint_url: str,
        cloud_creds: Optional[Dict[str, Any]] = None,
    ) -> "HLClient":
        transport = RequestsHTTPTransport(
            url=endpoint_url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Token {api_token}",
            },
            verify=True,  # SSL verification
            use_json=True,
        )
        client = GQLClient(
            transport=transport,
            fetch_schema_from_transport=True,
            execute_timeout=CONST_HLCLIENT_GQL_TIMEOUT_SEC,
        )
        return cls(client, cloud_creds=cloud_creds)

    def __repr__(self):
        return f"{self.endpoint_url}: [{self.api_token[:4]}...]"

    @classmethod
    def get_client(cls):
        """Instantiate a HLClient object as needed

        Singelton. If HLClient has already been instantiated this will return
        the existing instance, else will fallback on HLClient.get_client
        """
        if HLClient._instance is not None:
            return HLClient._instance
        return cls.from_env()

    @classmethod
    def clear_instance(cls):
        """Clear singleton instance, typically used in tests"""
        HLClient._instance = None

    @classmethod
    def from_env(cls) -> "HLClient":
        api_token = os.environ[ENV_HL_WEB_GRAPHQL_API_TOKEN]
        endpoint_url = os.environ[ENV_HL_WEB_GRAPHQL_ENDPOINT]
        cloud_creds = cls.cloud_creds_from_env()
        return cls.from_credential(api_token, endpoint_url, cloud_creds=cloud_creds)

    @classmethod
    def cloud_creds_from_env(cls) -> Optional[Dict[str, S3Creds]]:
        session = boto3.Session()
        session_credentials = session.get_credentials()
        aws_default_region = session.region_name
        aws_access_key_id = session_credentials.access_key if session_credentials is not None else None
        aws_secret_access_key = session_credentials.secret_key if session_credentials is not None else None

        if (
            (aws_default_region is not None)
            and (aws_access_key_id is not None)
            and (aws_secret_access_key is not None)
        ):
            cloud_creds = {
                S3Creds.__type__: S3Creds(
                    type=S3Creds.__type__,
                    aws_access_key_id=aws_access_key_id,
                    aws_secret_access_key=aws_secret_access_key,
                    aws_default_region=aws_default_region,
                )
            }
        else:
            cloud_creds = None
        return cloud_creds

    def export_credentials_to_environment(self):
        """Export credentials to environment variables."""
        os.environ[ENV_HL_WEB_GRAPHQL_API_TOKEN] = self.api_token
        os.environ[ENV_HL_WEB_GRAPHQL_ENDPOINT] = self.endpoint_url

        if self.cloud_creds is not None:
            for key, val in self.cloud_creds.items():
                if key == S3Creds.__type__:
                    for var_name, var in val.dict().items():
                        if var_name != "type":
                            os.environ[var_name.upper()] = var

    def append_credentials_to_env_file(self, outfile: str):
        """Append export KEY=VALUE lines for stored credentials"""
        token = self.api_token
        endpoint = self.endpoint_url

        # Token and Endpoint should always be present
        lines = [
            f"export {ENV_HL_WEB_GRAPHQL_API_TOKEN}={token}",
            f"export {ENV_HL_WEB_GRAPHQL_ENDPOINT}={endpoint}",
        ]

        # Cloud creds are optional. They are only needed if the client needs
        # to download data from a 3rd party's cloud bucket

        if self.cloud_creds is not None:
            for cloud_type, creds in self.cloud_creds.items():
                lines.extend(
                    [f"export {key}={value}" for key, value in creds.as_environment_variables().items()]
                )

        with open(str(outfile), "a+") as f:
            f.write("\n".join(lines))

    @property
    def endpoint_url(self):
        return self._client.transport.url

    @property
    def api_token(self):
        return self._client.transport.headers["Authorization"].split()[1]

    @property
    def account_name(self) -> str:
        return self.endpoint_url.split("//")[-1].split(".")[0]

    def get_s3_client(self):
        assert self.cloud_creds is not None
        assert S3Creds.__type__ in self.cloud_creds
        s3_creds = self.cloud_creds[S3Creds.__type__]
        s3_client = boto3.client(
            "s3",
            region_name=s3_creds.aws_default_region,
            aws_access_key_id=s3_creds.aws_access_key_id,
            aws_secret_access_key=s3_creds.aws_secret_access_key,
        )
        return s3_client

    def close(self):
        """Close the underlying GraphQL client and transport."""
        self._is_closed = True
        try:
            self._client.close()
        except Exception as e:
            LOGGER.warning(f"Error closing HLClient: {e}")

    def execute(self, request_str: str, variable_values=None) -> Dict[str, dict]:
        result = None

        LOGGER.debug(f"network_fn_decorator at call time: {decorators.network_fn_decorator}")

        def _execute(request_str: str, variable_vals=None):
            thread_id = threading.current_thread().ident
            thread_name = threading.current_thread().name

            execute_start = time.perf_counter()
            LOGGER.verbose(f"[Thread {thread_id}/{thread_name}] _execute: {request_str[:100]}...")

            gql_parse_start = time.perf_counter()
            request_gql = gql(request_str)
            gql_parse_elapsed = time.perf_counter() - gql_parse_start
            LOGGER.verbose(f"[Thread {thread_id}/{thread_name}] GQL parse took {gql_parse_elapsed:.3f}s")

            # Pre-serialize variable_vals using our custom JSON encoder to handle
            # UUIDs, datetimes, and other custom types, then parse back to dict
            if variable_vals is not None:
                serialized = HLJSONEncoder().encode(variable_vals)
                variable_vals = json.loads(serialized)

            client_execute_start = time.perf_counter()
            LOGGER.verbose(f"[Thread {thread_id}/{thread_name}] Using execute")

            result = self._client.execute(request_gql, variable_values=variable_vals)

            client_execute_elapsed = time.perf_counter() - client_execute_start
            LOGGER.verbose(
                f"[Thread {thread_id}/{thread_name}] GraphQL client.execute took {client_execute_elapsed:.3f}s"
            )

            execute_total = time.perf_counter() - execute_start
            LOGGER.verbose(f"[Thread {thread_id}/{thread_name}] Total _execute took {execute_total:.3f}s")
            return result

        decorator_start = time.perf_counter()
        result = decorators.network_fn_decorator(_execute)(request_str, variable_vals=variable_values)
        decorator_elapsed = time.perf_counter() - decorator_start
        LOGGER.verbose(f"Total execute (with decorator) took {decorator_elapsed:.3f}s")
        return result

    def __getstate__(self):
        state = dict(
            api_token=self.api_token,
            endpoint_url=self.endpoint_url,
            cloud_creds=self.cloud_creds,
        )
        return state

    def __setstate__(self, d):
        client = HLClient.from_credential(**d)
        self.__dict__ = client.__dict__.copy()

    def __getattr__(self, key) -> Callable[..., Any]:
        # methods = vars(type(self))
        # if key in methods:
        #     return methods[key]

        key = snake_to_camel(key)
        schema = self.schema
        if key not in get_all_queries(schema) + get_all_mutations(schema):
            raise ValueError(f"{key} is not a known query or mutation")

        def f(*, return_type: BaseModel, **kwargs):
            return_type_dict = get_gql_return_type(return_type)

            obj_type, target_gql_obj = get_gql_obj(schema, key)

            arg_lst = []
            arg_names = [x["name"] for x in target_gql_obj["args"]]
            for k in kwargs:
                if k not in arg_names:
                    raise ValueError(f"unknown gql argument '{k}' for '{key}'")

            for x in target_gql_obj["args"]:
                if x["name"] in kwargs:
                    arg_lst.append((x["name"], to_gql_type(x["type"])))

            indent = 0
            if len(arg_lst):
                lines = [
                    Line(line="%s _(" % obj_type.lower(), indent=indent),
                    *[Line(line=f"${x}: {t}", indent=indent + 1) for x, t in arg_lst],
                    Line(line=")", indent=indent),
                    Line(line="{", indent=indent),
                    Line(line=f"{key}(", indent=indent + 1),
                    *[Line(line=f"{x}: ${x}", indent=indent + 2) for x, _ in arg_lst],
                    Line(line=f")", indent=indent + 1),
                    *return_type_formatting(return_type_dict, indent + 1),
                    Line(line="}", indent=indent),
                ]
            else:
                lines = [
                    Line(line="%s _" % obj_type.lower(), indent=indent),
                    Line(line="{", indent=indent),
                    Line(line=f"{key}", indent=indent + 1),
                    *return_type_formatting(return_type_dict, indent + 1),
                    Line(line="}", indent=indent),
                ]

            generated_request_str = "\n".join([" " * 4 * x.indent + x.line for x in lines])
            LOGGER.verbose(f"executing {obj_type}:\n{generated_request_str}\nWith args {kwargs}")
            response = self.execute(generated_request_str, variable_values=kwargs)
            if "errors" in response:
                raise ValueError(response["errors"])

            result = response[key]
            try:
                return return_type(**result)
            except Exception as e:
                LOGGER.error(f"response: {response}")
                if getattr(return_type, "_name", None) == "List":
                    # for List[BaseModel]
                    return [return_type.__args__[0](**x) for x in result]
                raise e

        return f

    def hint(self, query_or_mutation: str):
        obj_type, target_gql_obj = get_gql_obj(self.schema, query_or_mutation)
        py_pos_arg_lst = []
        py_key_arg_lst = []
        for x in target_gql_obj["args"]:
            py_type = to_python_type(x["type"])
            if not py_type.startswith("Optional"):
                py_pos_arg_lst.append((x["name"], py_type))
            else:
                py_key_arg_lst.append((x["name"], py_type))

        args = [
            (
                Line(line=f"{n}: {t} = None,", indent=0)
                if t.startswith("Optional")
                else Line(line=cs.yellow(f"{n}: {t},"), indent=0)
            )
            for n, t in py_pos_arg_lst + py_key_arg_lst
        ]
        print(cs.red_black(query_or_mutation))
        print("\n".join([" " * 4 * x.indent + x.line for x in args]))


from threading import local
from weakref import WeakSet

_thread_local = local()
_thread_local_clients = WeakSet()


def close_all_thread_local_clients():
    """Close all tracked HLClient instances."""
    for client in list(_thread_local_clients):
        try:
            client.close()
        except Exception as exc:  # noqa: BLE001 - safety belt during shutdown
            LOGGER.warning("Failed to close HLClient during shutdown: %s", exc)


def get_threadsafe_hlclient(api_token, endpoint_url) -> "HLClient":
    """
    Return a *per-thread* HLClient.

    The first call in each thread creates a fresh HLClient (via from_env);
    subsequent calls re-use that same instance.  Different threads never
    share the underlying gql.Client, so it's safe with the `gql` library.
    """
    if hasattr(_thread_local, "hl_client"):
        if getattr(_thread_local.hl_client, "_is_closed", False):
            del _thread_local.hl_client

    if not hasattr(_thread_local, "hl_client"):
        # Pick whichever factory suits your deployment:
        #  - from_env()   : uses HL_* environment variables
        #  - from_profile('dev') : loads ~/.highlighter-profiles.yaml
        client = HLClient.from_credential(api_token, endpoint_url)
        _thread_local.hl_client = client
        _thread_local_clients.add(client)
    return _thread_local.hl_client



from typing import List, Optional, Tuple
from uuid import UUID

from pydantic import BaseModel

from highlighter.client.base_models.annotation import Annotation
from highlighter.client.base_models.observation import Observation
from highlighter.core import LabeledUUID
from highlighter.core.data_models.data_sample import DataSample


class Prediction(BaseModel):
    pixel_location: Tuple[float, float, float, float] | List[Tuple[float, float]] | None
    attribute: LabeledUUID
    value: LabeledUUID
    conf: float

    @property
    def points(self):
        if self.pixel_location is None:
            raise ValueError("Prediction has no pixel location")
        if isinstance(self.pixel_location, tuple):
            x0, y0, x1, y1 = self.pixel_location
            points = [(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)]
        else:
            points = list(self.pixel_location)
            if points and points[0] != points[-1]:
                points.append(points[0])
        return points

    def to_annotation(
        self,
        data_sample: DataSample,
        pipeline_element_name: Optional[str] = None,
    ) -> Annotation:
        """
        recorded_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))  # FIXME remove default
        stream_frame_index: int = 0
        media_frame_index: int = 0
        # TODO: Replace with data_source_id once we re-target annotations and submissions from data_files to data_sources
        data_file_id: Optional[UUID] = None

        """
        if self.pixel_location is None:
            raise ValueError("Cannot convert Prediction to Annotation if Prediction has no pixel location")

        ann = Annotation.from_points(
            self.points,
            self.conf,
            data_sample=data_sample,
        )
        self.add_to_annotation(ann)
        return ann

    def add_to_annotation(self, src_annotation: Annotation) -> Observation:
        obs = Observation.make_enum_observation(
            self.attribute,
            self.attribute.label,
            self.value.label,
            self.value,
            self.conf,
            src_annotation.occurred_at,
            pipeline_element_name=src_annotation.datum_source.pipeline_element_name,
            training_run_id=src_annotation.datum_source.training_run_id,
            host_id=src_annotation.datum_source.host_id,
            frame_id=src_annotation.datum_source.frame_id,
        )
        src_annotation.observations.add(obs)
        return obs

    def translate(self, xoff=0, yoff=0):
        if isinstance(self.pixel_location, tuple):
            x0, y0, x1, y1 = self.pixel_location
            self.pixel_location = (x0 + xoff, y0 + yoff, x1 + xoff, y1 + yoff)
        else:
            self.pixel_location = [(x + xoff, y + yoff) for x, y in self.pixel_location]
